"""holds the complex refactive index or permittivity data

material_data implements the Material class which can hold different
representations of spectral data (e.g. refractive index or permittivity).
The data is in the form of either a constant value, tabulated data or a model.
These different representations can be combined e.g. model for n (real part
of refractive index) and constant value for k (imaginary part of refractive
index)

Functions
---------
_check_table_shape
    validate that a numpy array has a given shape

Classes
-------
Material
    processes and interfaces refractive index data.
"""
from __future__ import print_function
import codecs
import numpy as np
#from dispersion import _str_to_class
import dispersion.spectral_data as spectral_data
from dispersion import Spectrum
from dispersion import Constant, Interpolation, \
                                                    Extrapolation

#from dispersion.spectral_data import _numeric_to_string_table
from dispersion.io import (Reader, _numeric_to_string_table,
                           _str_table_to_numeric)









def _check_table_shape(table, ncols, name):
    """
    check that numpy array shape has the correct number of columns
    """
    if (not len(table.shape) == 2 or
        not table.shape[1] == ncols):
        raise ValueError("tabulated {} data ".format(name) +
                         "must have shape Nx{}".format(ncols))

class Material():
    '''
    Class for processing refractive index and permittivity data

    Parameters
    ----------
    file_path: str
        file path from which to load data
    fixed_n: float
        fixed real part of refractive index
    fixed_nk: complex
        fixed complex refractive index
    fixed_eps_r: float
        fixed real part of permittivity
    fixed_eps: complex
        fixed complex permittivity
    tabulated_n: Nx2 array
        table of real part of refractive index to interpolate
    tabulated_nk: Nx3 array
        table of real and imaginary refractive index values to interpolate
    tabulated_n_alpha: Nx3 array
        table of real and refractive index and absorption coefficient values
        to interpolate
    tabulated_eps: Nx3 array
        table of real and imaginary permittivity values to interpolate
    model_kw: dict
        model parameters
    spectrum_type: str
        sets the default spectrum type
    unit: str
        sets the default unit
    meta_data: dict
        contains the meta data for the material
    data: dict
        holds one or two SpectralData objects to describe the data
    options: dict
        holds options for the material object
    defaults: dict
        default values for spectrum data

    Warnings
    --------
    the parameters file_path, fixed_n, fixed_nk, fixed_eps_r, fixed_eps,
    tabulated_n, tabulated_nk, tabulted_n_alpha, tabulated_eps and model_kw are mututally
    exclusive.

    '''

    def __init__(self, **kwargs):
        #parsing arguments
        parsed_args = self._parse_args(kwargs)
        #set inputs and defaults
        file_path = parsed_args["file_path"]
        #self.meta_data = None
        self.meta_data = {}
        self.meta_data['Reference'] = ""
        self.meta_data['Comment'] = ""
        self.meta_data['Name'] = ""
        self.meta_data['FullName'] = ""
        self.meta_data['Author'] = ""
        self.meta_data['Alias'] = ""
        self.meta_data['MetaComment'] = ""
        self.meta_data['Specification'] = {}
        self._file_data = None
        self.data = {'name': "",
                     'real': None,
                     'imag': None,
                     'complex':None}
        self.options = {'interp_oder':parsed_args["interp_order"]}
        self.defaults = {'unit':parsed_args["unit"],
                         'spectrum_type':parsed_args["spectrum_type"]}
        #process input arguments
        if file_path is not None:
            reader = Reader(file_path)
            file_data = reader.read_file()
            self._process_file_data(file_data)
        elif parsed_args['model_kw'] is not None:
            self._process_model_dict(parsed_args['model_kw'])
        elif parsed_args['tabulated_nk'] is not None:
            self._process_table(parsed_args['tabulated_nk'], 'nk')
        elif parsed_args['tabulated_n_alpha'] is not None:
            self._process_table(parsed_args['tabulated_n_alpha'], 'n_alpha')
        elif parsed_args['tabulated_n'] is not None:
            self._process_table(parsed_args['tabulated_n'], 'n')
        elif parsed_args['tabulated_eps'] is not None:
            self._process_table(parsed_args['tabulated_eps'], 'eps')
        else:
            self._process_fixed_value(parsed_args)

        self._complete_partial_data()

    def _parse_args(self, args):
        """
        validated the dictionary of class inputs
        """
        mutually_exclusive = {"file_path", "fixed_n", "fixed_nk",
                              "fixed_eps_r", "fixed_eps",
                              "tabulated_nk", "tabulated_n_alpha",
                              "tabulated_n", "tabulated_eps",
                              "model_kw"}
        inputs = {}
        n_mutually_exclusive = 0
        for arg in args.keys():
            if arg in args and args[arg] is not None:
                if arg in mutually_exclusive:
                    n_mutually_exclusive += 1
            inputs[arg] = args[arg]

        if n_mutually_exclusive == 0:
            raise ValueError("At least one of the following" +
                             " inputs is required: "+
                             "{}".format(mutually_exclusive))
        elif n_mutually_exclusive > 1:
            raise ValueError("Only one of the following" +
                             "inputs is allowed: "+
                             "{}".format(mutually_exclusive))
        # Check types
        str_args = {'file_path', 'spectrum_type', 'unit'}
        str_types = {str}
        self._check_type(inputs, str_args, str_types)
        if inputs['spectrum_type'] is None:
            inputs['spectrum_type'] = 'wavelength'
        if inputs['unit'] is None:
            inputs['unit'] = 'nanometer'
        if 'interp_order' not in inputs:
            inputs['interp_order'] = 1
        # pylint: disable=no-member
        # bug in pylint does not recognise numpy data types
        int_args = {'interp_oder'}
        int_types = {int}
        self._check_type(inputs, int_args, int_types)

        float_args = {"fixed_n", "fixed_eps_r"}
        float_types = {float, np.double}
        self._check_type(inputs, float_args, float_types)

        complex_args = {"fixed_nk", "fixed_eps"}
        complex_types = {complex, np.cdouble}
        self._check_type(inputs, complex_args, complex_types)

        dict_args = {'model_kw'}
        dict_types = {dict}
        self._check_type(inputs, dict_args, dict_types)

        array_args = {'tabulated_nk', 'tabulated_n', 'tabulated_eps',
                      'tabulated_n_alpha'}
        array_types = {np.ndarray}
        self._check_type(inputs, array_args, array_types)
        if inputs['tabulated_nk'] is not None:
            _check_table_shape(inputs['tabulated_nk'], 3, 'nk')
        if inputs['tabulated_n_alpha'] is not None:
            _check_table_shape(inputs['tabulated_nk'], 3, 'n_alpha')
        if inputs['tabulated_n'] is not None:
            _check_table_shape(inputs['tabulated_n'], 2, 'n')
        if inputs['tabulated_eps'] is not None:
            _check_table_shape(inputs['tabulated_eps'], 3, 'eps')

        return inputs

    @staticmethod
    def _check_type(args, names, types):
        """
        raises TypeError if the names keys in args dict are not in the
        set of types. If name is not in args, place a default value of None.
        """
        for arg in names:
            if arg in args and args[arg] is not None:
                invalid_type = False
                for _type in types:
                    if isinstance(args[arg], _type):
                        invalid_type = True
                if invalid_type is False:
                    raise TypeError("argument " +
                                    "{} must be".format(arg) +
                                    " of types: {}".format(types))
            else:
                args[arg] = None


    def _complete_partial_data(self):
        """
        if only partial data was provided then set remaining parameters
        to constant value of 0.
        """
        if self.data['real'] is None:
            self.data['real'] = Constant(0.0)
        if self.data['imag'] is None:
            self.data['imag'] = Constant(0.0)

    def remove_absorption(self):
        """
        sets loss (k or epsi) to constant zero value

        Warnings
        --------
        has no effect if the material is defined as via complex data instead of
        separate real and imaginary parts.
        """
        self.data['imag'] = Constant(0.0)

    def extrapolate(self, new_spectrum, spline_order=2):
        """extrpolates the material data

        extrapolates the material data to cover the range defined by the
        spectrum new_spectrum. if new_spectrum has only one element, the data
        will be extrapolated from the relevant end of its valid range up to the
        value given by new_spectrum. spline_order defines the order of the
        spline used for extrapolation. The results of the extrapolation depend
        heavily on the order chosen, so please check the end result to make
        sure it make physical sense.

        Parameters
        ----------
        new_spectrum: Spectrum
            the values to exrapolate to
        spline_order: int
            the order of spline to use for interpolation -> extrpolation

        Raises
        ------
        NotImplementedError
            if the material is defined as via a complex value

        """

        if self.data['complex'] is None:
            for data_name in ['real', 'imag']:
                if isinstance(self.data[data_name], Constant):
                    continue
                self.data[data_name] = Extrapolation(self.data[data_name],
                                                     new_spectrum,
                                                     spline_order=spline_order)
        else:
            raise NotImplementedError("extrapolation not implemented " +
                                      "for materials with real and imaginary "+
                                      "parts not independent from each other")

    def _process_fixed_value(self, inputs):
        '''use fixed value inputs to set n/k or permittivity

        the fixed value is converted to a SpectralData.Constant object and
        included in the data dict

        Parameters
        ----------
        inputs: dict
            the dict holding the fixed value
        '''
        if inputs['fixed_n'] is not None:
            self.data['name'] = 'nk'
            self.data['real'] = Constant(inputs['fixed_n'])
            #self._k = Constant(0.0)
        elif inputs['fixed_nk'] is not None:
            self.data['name'] = 'nk'
            self.data['real'] = Constant(np.real(inputs['fixed_nk']))
            self.data['imag'] = Constant(np.imag(inputs['fixed_nk']))
        elif inputs['fixed_eps_r'] is not None:
            self.data['name'] = 'eps'
            self.data['real'] = Constant(inputs['fixed_eps_r'])
            #self._epsi = Constant(0.0)
        elif inputs['fixed_eps'] is not None:
            self.data['name'] = 'eps'
            self.data['real'] = Constant(np.real(inputs['fixed_eps']))
            self.data['imag'] = Constant(np.imag(inputs['fixed_eps']))
        else:
            raise RuntimeError("Failed to set a constant value for n,k or eps")

    def _process_model_dict(self, model_dict):
        """use model parameter input to set n/k or permittivity

        use model_dict to return a SpectralData.Model object and sets the
        relevant n/k or permittivity class attributes

        Parameters
        ----------
        model_dict: dict
            contains data for model creates (see notes)

        Raises
        ------
        ValueError
            if the model output does not yield n/k or permittivity

        Notes
        -----
        model_dict must contain the fields:
        name: str
            class name of the model (see spectral_data.py)
        spectrum_type: str
            spectrum Type (see spctrum.py)
        unit: str
            spetrum unit (see spctrum.py)
        valid_range: 2x1 np.array
            min and max of the spectral range for which the model is valid
        parameters: np.array
            all paramters (i.e. coefficients) needed for the model
        """
        model_class = self._str_to_class(model_dict['name'])
        #model_class = MODELS[model_dict['name']]
        kws = {}
        if "spectrum_type" in model_dict:
            kws['spectrum_type'] = model_dict['spectrum_type']
            self.defaults['spectrum_type'] = model_dict['spectrum_type']
        if "unit" in model_dict:
            kws['unit'] = model_dict['unit']
            self.defaults['unit'] = model_dict['unit']
        model = model_class(model_dict['parameters'],
                            model_dict['valid_range'], **kws)
        if model.output == 'n':
            self.data['name'] = 'nk'
            self.data['real'] = model
        elif model.output == 'k':
            self.data['name'] = 'nk'
            self.data['imag'] = model
        elif model.output == 'nk':
            self.data['name'] = 'nk'
            self.data['complex'] = model
        elif model.output == 'epsr':
            self.data['name'] = 'eps'
            self.data['real'] = model
        elif model.output == 'epsi':
            self.data['name'] = 'eps'
            self.data['imag'] = model
        elif model.output == 'eps':
            self.data['name'] = 'eps'
            self.data['complex'] = model
        else:
            raise ValueError("model output <{}> invalid".format(model.output))

    @staticmethod
    def _str_to_class(field):
        """evaluates string as a class.

        tries to evaluate the given string as a class from the spectral_data
        module.

        Parameters
        ----------
        field: str
            name to convert to class

        Raises
        ------
        NameError
            the given field is not an attribute
        TypeError
            the given field is an attribute but not a class
        """
        try:
            identifier = getattr(spectral_data, field)
        except AttributeError:
            raise NameError("%s doesn't exist." % field)
        if isinstance(identifier, type):
            return identifier
        raise TypeError("%s is not a class." % field)

    def _process_file_data(self, file_dict):
        """set meta_data and data from dictionary"""
        self._file_data = file_dict
        self.meta_data = {}
        self.meta_data['Reference'] = file_dict['MetaData']['Reference']
        self.meta_data['Comment'] = file_dict['MetaData']['Comment']
        self.meta_data['Name'] = file_dict['MetaData']['Name']
        self.meta_data['FullName'] = file_dict['MetaData']['FullName']
        self.meta_data['Author'] = file_dict['MetaData']['Author']
        self.meta_data['MetaComment'] = file_dict['MetaData']['MetaComment']
        self.meta_data['Specification'] = file_dict['MetaData']['Specification']
        datasets = file_dict['Datasets']
        #self.dataTypes = []
        #self.dataSets = []
        for dataset in datasets:
            data_type, identifier = dataset['DataType'].split()
            #meta_data = dataset['MetaData']
            if data_type == 'tabulated':
                #data is tabulated
                dataset['Data'] = _str_table_to_numeric(dataset.pop('Data'))
                self._process_table(dataset['Data'], identifier,
                                    meta_data=dataset)
            elif data_type in {'formula', 'model'}:
                #data is a formula with coefficients
                self._process_formula_data(dataset)
            else:
                raise ValueError("data type {} not supported".format(data_type))

    def _process_table(self, table, identifier, meta_data=None):
        """
        Uses a table(np.ndarray) and metadata to set relevant class attributes.
        """
        if meta_data is None:
            meta_data = {}
        if ('SpectrumType' in meta_data and \
            meta_data['SpectrumType'] and \
            meta_data['SpectrumType'] is not None):
            self.defaults['spectrum_type'] = meta_data['SpectrumType']
        if ('Unit' in meta_data and \
            meta_data['Unit'] and \
            meta_data['Unit'] is not None):
            self.defaults['unit'] = meta_data['Unit']
        if identifier == 'nk':
            self.data['name'] = 'nk'
            self.data['real'] = self._spec_data_from_table(table[:, [0, 1]])
            self.data['imag'] = self._spec_data_from_table(table[:, [0, 2]])
        elif identifier == 'n_alpha':
            self.data['name'] = 'nk'
            self.data['real'] = self._spec_data_from_table(table[:, [0, 1]])
            default_spectrum = Spectrum(table[:, 0],
                                        self.defaults['spectrum_type'],
                                        self.defaults['unit'])
            wvls = default_spectrum.convert_to('wavelength','cm')
            alpha = table[:, 2]
            k = alpha*wvls/(np.pi*4.)
            new_table = np.vstack([table[:, 0], k]).T
            self.data['imag'] = self._spec_data_from_table(new_table)
        elif identifier == 'n':
            self.data['name'] = 'nk'
            self.data['real'] = self._spec_data_from_table(table)
        elif identifier == 'k':
            self.data['name'] = 'nk'
            self.data['imag'] = self._spec_data_from_table(table)
        elif identifier == 'eps':
            self.data['name'] = 'eps'
            self.data['real'] = self._spec_data_from_table(table[:, [0, 1]])
            self.data['imag'] = self._spec_data_from_table(table[:, [0, 2]])


    def _spec_data_from_table(self, data):
        '''
        Convert table to SpectralData object.

        Parameters
        ----------
        data: Nx2 np.array
            the tabulated spectral data

        Returns
        -------
        Constant(SpectralData)
            if tabulated data has 1 row the data is constant
        Interpolation(SpectralData)
            interpolation of the tabulated data
        '''
        n_rows = data.shape[0]
        spec_type = self.defaults['spectrum_type']
        unit = self.defaults['unit']
        if n_rows == 1:
            return Constant(data[0, 1],
                            valid_range=(data[0, 0], data[0, 0]),
                            spectrum_type=spec_type, unit=unit)

        return Interpolation(data, spectrum_type=spec_type,
                             unit=unit)

    def _process_formula_data(self, data_dict):
        '''prepare dictionary of data for processing.

        create model_dict and call process_model_dict use range and coefficients
        in input dictionary to return a SpectralData.Model
        '''
        model_dict = {}
        meta_data = data_dict
        data_type, identifier = meta_data['DataType'].split()
        if not (data_type in {'formula', 'model'}):
            raise ValueError("dataType <{}>".format(data_type) +
                             " not a valid formula or model")
        if data_type == 'formula':
            identifier = int(identifier)

        if meta_data['ValidRange']:
            valid_range = meta_data['ValidRange'].split()

        for i_valid_range, v_range in enumerate(valid_range):
            valid_range[i_valid_range] = float(v_range)
        model_dict['valid_range'] = valid_range

        coefficients = data_dict['Data'].split()
        for iter_coeff, coeff in enumerate(coefficients):
            coefficients[iter_coeff] = float(coeff)
        model_dict['parameters'] = np.array(coefficients)


        if meta_data['SpectrumType']:
            model_dict['spectrum_type'] = meta_data['SpectrumType']
        else:
            model_dict['spectrum_type'] = self.defaults['spectrum_type']

        if meta_data['Unit']:
            model_dict['unit'] = meta_data['Unit']
        else:
            model_dict['unit'] = self.defaults['unit']

        method_ids = {1: 'Sellmeier', 2: 'Sellmeier2',
                      3: 'Polynomial', 4: 'RefractiveIndexInfo',
                      5: 'Cauchy', 6: 'Gases',
                      7: 'Herzberger', 8: 'Retro',
                      9: 'Exotic'}

        if isinstance(identifier, int):
            model_dict['name'] = method_ids[identifier]
        else:
            model_dict['name'] = identifier

        self._process_model_dict(model_dict)


    def get_nk_data(self, spectrum,
                    spectrum_type='wavelength',
                    unit='meter'):
        '''
        return complex refractive index for a given input spectrum.

        Parameters
        ----------
        spectrum: np.array or Spectrum
            the spectral values to evaluate
        spectrum_type: str {'wavelength', 'frequency', 'energy'}
            type of spectrum
        unit: str {'meter', 'nanometer', 'micrometer', 'hertz', 'electronvolt'}
            unit of spectrum (must match spectrum type)

        Returns
        -------
        np.complex128
            the complex n/k values (if input spectrum has size == 1)
        np.array with np.complex128 dtype
            the complex n/k values (if input spectrum has size > 1)
        '''
        if isinstance(spectrum, Spectrum):
            spectrum_values = spectrum.values
            spectrum_type = spectrum.spectrum_type
            unit = spectrum.unit
        else:
            spectrum_values = spectrum

        spectrum = Spectrum(spectrum_values,
                            spectrum_type=spectrum_type,
                            unit=unit)
        valid_data_names = ['nk', 'eps']
        if not (self.data['name'] in valid_data_names):
            raise ValueError("data type {}".format(self.data['name']) +
                             "cannot be converted to refractive index")
        if self.data['complex'] is None:
            real = self.data['real'].evaluate(spectrum)
            imag = 1j*self.data['imag'].evaluate(spectrum)
            complex_val = real+imag
        else:
            complex_val = self.data['complex'].evaluate(spectrum)

        if self.data['name'] == 'eps':
            complex_val = np.sqrt(complex_val)
        return complex_val

    def get_permittivity(self, spectrum_values,
                         spectrum_type='wavelength',
                         unit='meter'):
        '''
        return complex permittivity for a given input spectrum.

        Parameters
        ----------
        spectrum: np.array or Spectrum
            the spectral values to evaluate
        spectrum_type: str {'wavelength', 'frequency', 'energy'}
            type of spectrum
        unit: str {'meter', 'nanometer', 'micrometer', 'hertz', 'electronvolt'}
            unit of spectrum (must match spectrum type)

        Returns
        -------
        np.complex128
            the complex permittivity values (if input spectrum has size == 1)
        np.array with np.complex128 dtype
            the complex permittivity values (if input spectrum has size > 1)
        '''

        if isinstance(spectrum_values, Spectrum):
            spectrum = spectrum_values
        else:
            spectrum = Spectrum(spectrum_values,
                                spectrum_type=spectrum_type,
                                unit=unit)

        if not (self.data['name'] == 'nk' or self.data['name'] == 'eps'):
            raise ValueError("data type {}".format(self.data['name']) +
                             "cannot be converted to refractive index")

        if self.data['complex'] is None:
            real = self.data['real'].evaluate(spectrum)
            imag = 1j*self.data['imag'].evaluate(spectrum)
            complex_val = real+imag
        else:
            complex_val = self.data['complex'].evaluate(spectrum)

        if self.data['name'] == 'nk':
            complex_val = np.power(complex_val, 2)
        return complex_val

    def get_maximum_valid_range(self):
        """find maximum spectral range that spans real and imaginary data.

        Checks both real and imaginary parts of spectral data and finds the
        maximum spectral range which is valid for both parts.

        Returns
        -------
        2x1 np.array
            the maximum valid range
        """
        if not(self.data['name'] == 'nk' or self.data['name'] == 'eps'):
            raise RuntimeError("valid_range cannot be defined as "+
                               "Material does not yet contain "+
                               " a valid n/k or permittivity spectrum")

        if self.data['complex'] is None:
            real_range_std = self.data['real'].valid_range.standard_rep
            imag_range_std = self.data['imag'].valid_range.standard_rep
            real_lower = np.min(real_range_std)
            real_upper = np.max(real_range_std)
            imag_lower = np.min(imag_range_std)
            imag_upper = np.max(imag_range_std)
            lower = np.max([real_lower, imag_lower])
            upper = np.min([real_upper, imag_upper])
        else:
            lower = np.min(self.data['complex'].valid_range.values)
            upper = np.max(self.data['complex'].valid_range.values)
        max_range = np.array([lower, upper])
        spec = Spectrum(max_range)
        return spec.convert_to(self.defaults['spectrum_type'],
                               self.defaults['unit'])

    @staticmethod
    def utf8_to_ascii(string):
        """converts a string from utf8 to ascii"""
        uni_str = codecs.encode(string, 'utf-8')
        ascii_str = codecs.decode(uni_str, 'ascii', 'ignore')
        return ascii_str

    def print_reference(self):
        """print material reference"""
        print(self.utf8_to_ascii(self.meta_data['Reference']))

    def print_comment(self):
        """print material comment"""
        print(self.utf8_to_ascii(self.meta_data['Comment']))


    def plot_nk_data(self, **kwargs):
        """plots the real and imaginary part of the refractive index"""
        self._plot_data('nk', **kwargs)

    def plot_permittivity(self, **kwargs):
        """plots the real and imaginary part of the permittivity"""
        self._plot_data('permittivity', **kwargs)

    def _plot_data(self, data_label, **kwargs):
        """internal function used for plotting spectral data"""
        try:
            import matplotlib.pyplot as plt
        except ModuleNotFoundError as exp:
            raise ModuleNotFoundError("plotting requires the matplotlib",
                                      " package to be installed")

        plot_data = self._prepare_plot_data(**kwargs)
        if 'axes' not in kwargs:
            plot_data['axes'] = plt.axes()
        else:
            plot_data['axes'] = kwargs['axes']
        if data_label == 'nk':
            data = self.get_nk_data(plot_data['spectrum'])
            labels = ['n', 'k']
        elif data_label == 'permittivity':
            data = self.get_permittivity(plot_data['spectrum'])
            labels = ['eps_r', 'eps_i']

        data_r = np.real(data)
        data_i = np.imag(data)
        # pylint: disable=protected-access
        # this is the only way to access the color cycler
        axes = plot_data['axes']
        spectrum = plot_data['spectrum']
        if spectrum.values.size == 1:
            color = next(axes._get_lines.prop_cycler)['color']
            plt.axhline(data_r, label=labels[0], color=color)
            color = next(axes._get_lines.prop_cycler)['color']
            plt.axhline(data_i, label=labels[1], ls='--', color=color)
        else:
            plt.plot(spectrum.values, data_r, label=labels[0])
            plt.plot(spectrum.values, data_i, ls='--', label=labels[1])
        plt.legend(loc='best')
        plt.ylabel("{}, {}".format(labels[0], labels[1]))
        xlabel = spectrum.get_type_unit_string()
        plt.xlabel(xlabel)

    def _prepare_plot_data(self, **kwargs):
        """internal function to prepare data for plotting"""
        plot_data = {}
        if 'spectrum_type' not in kwargs:
            plot_data['spectrum_type'] = self.defaults['spectrum_type']
        else:
            plot_data['spectrum_type'] = kwargs['spectrum_type']
        if 'unit' not in kwargs:
            plot_data['unit'] = self.defaults['unit']
        else:
            plot_data['unit'] = kwargs['unit']
        if 'values' not in kwargs:
            spectrum = self.get_sample_spectrum()
            values = spectrum.convert_to(plot_data['spectrum_type'],
                                         plot_data['unit'])


        else:
            values = kwargs['values']
            if isinstance(values, (list, tuple)):
                values = np.array(values)
        spectrum = Spectrum(values,
                            spectrum_type=plot_data['spectrum_type'],
                            unit=plot_data['unit'])
        plot_data['spectrum'] = spectrum
        return plot_data

    def get_sample_spectrum(self):
        """spectrum which covers the maximum valid range of the material data"""
        max_range = self.get_maximum_valid_range()
        if max_range[0] == 0.0 or max_range[1] == np.inf:
            values = np.geomspace(100, 2000, 1000)
            spectrum = Spectrum(values, spectrum_type='wavelength',
                                unit='nm')
        else:
            values = np.geomspace(max_range[0], max_range[1], 1000)
            if values[0] < max_range[0]:
                values[0] = max_range[0]
            if values[-1] > max_range[1]:
                values[-1] = max_range[1]
            spectrum = Spectrum(values,
                                spectrum_type=self.defaults['spectrum_type'],
                                unit=self.defaults['unit'])
        return spectrum

    def prepare_file_dict(self):
        #if self._file_data is not None:
        #    return self._file_data
        file_dict = {}
        file_dict['MetaData'] = {}
        #file_dict[]
        for key in self.meta_data:
            if key == "Alias":
                continue
            file_dict['MetaData'][key] = self.meta_data[key]
        file_dict['Datasets'] = self.dataset_to_dict()
        return file_dict

    def add_dtype_suffix(self, dtype, data_part):
        if self.data['name'] == 'nk':
            if data_part == 'real':
                dtype += " n"
            elif data_part == 'imag':
                dtype += " k"
            elif data_part == 'complex':
                dtype += ' nk'
        elif self.data['name'] == 'eps':
            if data_part == 'real':
                dtype += " eps_r"
            elif data_part == 'imag':
                dtype += ' eps_k'
            elif data_part == 'complex':
                dtype += ' eps'
        else:
            raise ValueError("data name could not be parsed")
        return dtype


    def dataset_to_dict(self):
        """
        generate a file_data type dictionary from this object

        Parameters
        ----------

        material_data: dict
            keys: name, real, imag, complex

        Returns
        -------
        dict
            a list of dicts that has a format suitable for writing to file
        """
        datasets = []
        if self.data['complex'] is None:
            data_parts = ['real', 'imag']
        else:
            data_parts = 'complex'


        for data_part in data_parts:
            spec_data = self.data[data_part]
            data_dict = spec_data.dict_repr()
            if isinstance(spec_data, (Constant, Interpolation)):
                dtype = data_dict['DataType']
                dtype = self.add_dtype_suffix(dtype, data_part)
                data_dict['DataType'] = dtype

            datasets.append(data_dict)
        datasets = self.collapse_datasets(datasets)
        return datasets

    def collapse_datasets(self, datasets):
        n_collapsable = 0
        for dataset in datasets:
            if dataset['DataType'] == 'tabulated n':
                n_data = _str_table_to_numeric(dataset['Data'])
                n_collapsable += 1
            if dataset['DataType'] == 'tabulated k':
                k_data = _str_table_to_numeric(dataset['Data'])
                n_collapsable += 1

        if n_collapsable < 2:
            return datasets

        collapse = True
        if not np.all(n_data[:, 0] == k_data[:, 0]):
            collapse = False

        if datasets[0]['Unit'] != datasets[1]['Unit']:
            collapse = False

        if datasets[0]['SpectrumType'] != datasets[1]['SpectrumType']:
            collapse = False

        if not collapse:
            return datasets

        new_dataset = {}
        new_dataset['Unit'] = datasets[0]['Unit']
        new_dataset['SpectrumType'] = datasets[0]['SpectrumType']
        new_dataset['DataType'] = 'tabulated nk'
        k_data = k_data[:, 1].reshape(k_data.shape[0], 1)
        new_data = np.concatenate([n_data, k_data], axis=1)
        new_dataset['Data'] = _numeric_to_string_table(new_data)
        return [new_dataset]

class EffectiveMedium(Material):

    def __init__(self, spectrum, material1, material2, filling_fraction):
        self.data = {'name': "",
                     'real': None,
                     'imag': None,
                     'complex':None}
        self.options = {'interp_oder':1}
        self.defaults = {'unit':'m',
                         'spectrum_type':'wavelength'}
        self.spectrum = spectrum
        self.mat1 = material1
        self.mat2 = material2
        self.frac = filling_fraction
        if self.frac < 0. or self.frac > 1.0:
            raise ValueError("filling fraction must be between "+
                             "0. and 1.")

        self.create_effective_data()

    def create_effective_data(self):
        """
        this must be implemented in a subclass
        """
        raise NotImplementedError("create_effective_data must be defined" +
                                  " in a subclass")

class MaxwellGarnett(EffectiveMedium):

    def create_effective_data(self):
        #def get_maxwell_garnet(eps_base, eps_incl, vol_incl):
        small_number_cutoff = 1e-6
        eps_base = self.mat1.get_permittivity(self.spectrum)
        eps_incl = self.mat2.get_permittivity(self.spectrum)
        factor_up = 2*(1-self.frac)*eps_base+(1+2*self.frac)*eps_incl
        factor_down = (2+self.frac)*eps_base+(1-self.frac)*eps_incl
        if np.any(abs(factor_down)) < small_number_cutoff:
            raise ValueError('effective medium is approximately singular')
        eps_eff = eps_base*factor_up/factor_down
        self.spectrum.convert_to("wavelength", 'm', in_place=True)
        table = np.vstack([[self.spectrum.values],
                                [np.real(eps_eff)],
                                [np.imag(eps_eff)]]).T
        self._process_table(table, "eps")

class Bruggeman(EffectiveMedium):

    def create_effective_data(self):
        eps_b = self.mat1.get_permittivity(self.spectrum)
        eps_a = self.mat2.get_permittivity(self.spectrum)
        solution1 = self._get_bruggeman_solution1(eps_b, eps_a)
        solution2 = self._get_bruggeman_solution2(eps_b, eps_a)
        self.sol1 = solution1
        self.sol2 = solution2
        indices1 = np.imag(solution1) >= 0.0
        indices2 = np.imag(solution2) >= 0.0
        eps_eff = np.zeros(eps_b.shape, dtype=np.cdouble)
        eps_eff[indices2] = solution2[indices2]
        eps_eff[indices1] = solution1[indices1]
        self.spectrum.convert_to("wavelength", 'm', in_place=True)
        table = np.vstack([[self.spectrum.values],
                                [np.real(eps_eff)],
                                [np.imag(eps_eff)]]).T
        self._process_table(table, "eps")

    def _get_bruggeman_solution1(self, eps_b, eps_a):
        q = -0.5*eps_a*eps_b
        p = 0.5*(self.frac*(eps_b-2*eps_a) + (1-self.frac)*(eps_a-2*eps_b) )
        return -p*0.5  + np.sqrt((0.5*p)**2 - q)

    def _get_bruggeman_solution2(self, eps_b, eps_a):
        q = -0.5*eps_a*eps_b
        p = 0.5*(self.frac*(eps_b-2*eps_a) + (1-self.frac)*(eps_a-2*eps_b) )
        return -p*0.5  - np.sqrt((0.5*p)**2 - q)







if __name__ == "__main__":
    pass



"""A database for administering refractive index data files

The database set of data files on disk which describe spectrally resolved
refractive index or permittivity data

Functions
---------

rebuild_database
    rebuilds all of the modules in the database

Classes
-------

Catalogue
    administers the catalogue of material data files
"""
from __future__ import print_function
import os
import warnings
import numpy as np
import pandas as pd
ver = pd.__version__
split = ver.split(".")
PANDAS_MINOR_VERSION = int(split[1])

from dispersion.material import Material
from dispersion.spectrum import Spectrum
from dispersion.config import get_config, validate_config
from dispersion.io import read_yaml_file

def rebuild_catalogue():
    valid_ans = False
    while not valid_ans:
        ans = input("really rebuild and overwrite material catalogue? (y/n) ")
        if ans == 'n':
            valid_ans = True
            return
        elif ans == 'y':
            valid_ans = True

    cat = Catalogue(rebuild='All')

    cat.save_to_file()


class Catalogue(object):
    """
    administers the set of material data files

    this class is used to find and load data files from disk which describe
    spectrally resolved refractive index or permittivity data into Material
    objects.

    Parameters
    ----------

    config: dict or None
        configuration data
    rebuild: list or None
        which modules which should be rebuilt
    base_path: str
        where the database file structure is stored
    file_name: str
        name of the catalogue file
    database: pandas.DataFrame
        the catalogue of material files
    qgrid_widget: qgrid.widget
        iPython widget for interactive editing of the catalogue
    make_grid: function
        qgrid function for refreshing the interactive interface
    reference_spectrum: Spectrum
        catalogue will provide n/k values at the reference spectrum value
    rii_loader: dict
        temporary dict used in constructing the refractive index info catalogue
    """
    # pylint: disable=no-member
    # bug in pylint does not recognise numpy data types

    META_DATA = {'Alias':str,
                 'Name':str,
                 'FullName':str,
                 'Author':str,
                 'Comment':str,
                 'Reference':str,
                 'SpectrumType':str,
                 'Unit':str,
                 'SpectrumLowerBound':np.double,
                 'SpectrumUpperBound':np.double,
                 'N_Reference':np.double,
                 'K_Reference':np.double,
                 'Path':str,
                 'Module':str}
    NA_VALUES = {'N_Reference':[""],
                 'K_Reference':[""]}


    def __init__(self, config=None, rebuild="None"):
        if config is None:
            config = get_config()
        validate_config(config)
        self.make_reference_spectrum(config)
        self.config = config
        self.base_path = config['Path']
        self.file_name = config['File']
        if rebuild == 'All':
            df = pd.DataFrame(columns=Catalogue.META_DATA.keys())
        else:
            dtypes = Catalogue.META_DATA
            df = pd.read_csv(os.path.join(self.base_path,
                                          self.file_name),
                             dtype=dtypes,
                             na_values=Catalogue.NA_VALUES,
                             keep_default_na=False)

        if rebuild != 'None':
            df = self.build_catalogue(df, rebuild)
        self.database = df
        self.qgrid_widget = None
        if self.config['Interactive']:
            try:
                import qgrid
            except ModuleNotFoundError as exc:
                raise Exception("Interactive mode requires the package" +
                                ' qgrid') from exc
            self.make_qgrid = qgrid.show_grid

    def build_catalogue(self, df, rebuild):
        """
        read specified modules and return a dataframe combining all modules
        """
        config_modules = self.config['Modules']
        all_modules = {'RefractiveIndexInfo':self.read_ri_info_db,
                       'Filmetrics':self.read_filmetrics_db,
                       'UserData':self.read_user_data_db}
        df_new = pd.DataFrame(columns=Catalogue.META_DATA.keys())
        for module, valid in config_modules.items():
            if valid:
                if module in rebuild or rebuild == 'All':
                    print("Building {}".format(module))
                    db_path = module
                    dir_path = os.path.join(self.base_path, db_path)
                    read_function = all_modules[module]
                    dframe = read_function(dir_path)
                    if PANDAS_MINOR_VERSION > 22:
                        df_new = df_new.append(dframe, sort=False,
                                               ignore_index=True)
                    else:
                        df_new = df_new.append(dframe, ignore_index=True)
                else:
                    dframe = df[df.Module == module]
                    if PANDAS_MINOR_VERSION > 22:
                        df_new = df_new.append(dframe, sort=False,
                                               ignore_index=True)
                    else:
                        df_new = df_new.append(dframe, ignore_index=True)

            else:
                if rebuild == module:
                    raise ValueError("Tried to rebuild module" +
                                     "{}".format(rebuild) +
                                     ", however this module is disabled" +
                                     " in the configuration.")
        return df_new

    def view_interactive(self):
        """
        returns a read only qgrid instance for interactively viewing the
        catalogue
        """
        if self.config['Interactive'] is False:
            raise ValueError("interactivity disabled in config")
        grid_options = {'editable':False}
        self.qgrid_widget = self.make_qgrid(self.database,
                                            show_toolbar=True,
                                            precision=3,
                                            grid_options=grid_options)
        return self.qgrid_widget


    def edit_interactive(self):
        """
        returns an editable qgrid instance for interactively viewing the
        catalogue. Call the method save_interactive to save any changes made
        """
        if self.config['Interactive'] is False:
            raise ValueError("interactivity disabled in config")
        grid_options = {'editable':True}
        self.qgrid_widget = self.make_qgrid(self.database,
                                            show_toolbar=True,
                                            precision=3,
                                            grid_options=grid_options)
        return self.qgrid_widget

    def save_interactive(self):
        """
        saves changes made to the qgrid when interfactive edit mode has been
        used"""
        if self.config['Interactive'] is False:
            raise ValueError("interactivity disabled in config")
        self.database = self.qgrid_widget.get_changed_df()

    def get_database(self):
        """returns the pandas data frame"""
        return self.database

    def set_database(self, database):
        """set the pandas data frame"""
        self.database = database

    def save_to_file(self):
        """save the pandas dataframe to the root path of the catalogue
        file structure"""
        self.database.to_csv(os.path.join(self.base_path, self.file_name),
                             index=False, index_label='Index')

    def register_alias(self, row_id, alias):
        """create an alias for a material to easily access it from the
        catalogue

        Parameters:
        rows_id: int or sliced DataFrame
            the row index or row of a df in which to set the alias
        alias: str
            the alias to set
        """
        index = None
        if isinstance(row_id, pd.core.frame.DataFrame):
            index = row_id.index
        elif isinstance(row_id, int):
            index = row_id
        if index is None:
            raise ValueError("row_id: {}".format(row_id) +
                             " with type {}".format(type(row_id)) +
                             " not understood")
        if alias in self.database.loc[:, 'Alias'].values:
            if not self.database.at[index, 'Alias'] == alias:
                raise ValueError("Alias {} ".format(alias) +
                                 "already in use. Failed to " +
                                 "add to catalogue.")

        self.database.at[index, 'Alias'] = alias

    def get_material(self, identifier):
        """get a material from the catalogue using its alias or row number"""
        if isinstance(identifier, str):
            bool_array = self.database.Alias.values == identifier
            bool_list = bool_array.tolist()
            row_index = self.database.index[bool_list]
            if row_index.size == 0:
                raise ValueError("identifier {} does not ".format(identifier) +
                                 "name a valid alias in the " +
                                 "catalogue")
            row = self.database.iloc[row_index[0], :]

        elif isinstance(identifier, int):
            row = self.database.iloc[identifier, :]
        else:
            raise ValueError("identifier must be of type str or int")
        file_path = os.path.normpath(os.path.join(self.base_path,
                                                  row.Module,
                                                  row.Path.replace('\\', '/')))
        mat = Material(file_path=file_path,
                           spectrum_type=row.SpectrumType,
                           unit=row.Unit)
        return mat

    def make_reference_spectrum(self, config):
        """make the spectrum with which every material in the catalogue will
        be evaluated"""
        val = config['ReferenceSpectrum']['Value']
        spec_type = config['ReferenceSpectrum']['SpectrumType']
        unit = config['ReferenceSpectrum']['Unit']
        self.reference_spectrum = Spectrum(val, spectrum_type=spec_type,
                                           unit=unit)


    def read_ri_info_db(self, db_path):
        """read the file structure provided by the refractiveindex.info
        website"""
        self.rii_loader = {}
        self.rii_loader['db_path'] = db_path
        data = read_yaml_file(os.path.join(db_path, "database", "library.yml"))
        dframe = pd.DataFrame(columns=Catalogue.META_DATA.keys())
        self.rii_loader['database_list'] = []
        self._iterate_shelves(data)

        dframe = pd.DataFrame(self.rii_loader['database_list'],
                              columns=Catalogue.META_DATA.keys())
        self.rii_loader = None
        return dframe

    def _iterate_shelves(self, shelves):
        """
        shelves is an ordered dict
        """
        for shelf in shelves:
            #shelf = shelf_data['SHELF']
            self.rii_loader['current_shelf'] = shelf['name']
            books = shelf['content']
            self._iterate_books(books)


    def _iterate_books(self, books):
        """
        iterate through the books on the shelf
        """
        for book in books:
            if "DIVIDER" in book.keys():
                self.rii_loader['current_divider'] = book['DIVIDER']
                continue
            elif "BOOK" in book.keys():
                self.rii_loader['current_book'] = book['BOOK']
                self.rii_loader['current_full_name'] = book['name']
                pages = book['content']
                self._iterate_pages(pages)


    def _iterate_pages(self, pages):
        """
        iterate the pages of the book and add to catalogue

        The pages of the refactiveindex.info catalogue are data files. We iterate
        over them and add the current shelf, book and page to the catalogue. A
        Material object is created from the data file.
        """
        for page in pages:
            if "DIVIDER" in page:
                continue
            elif "PAGE" in page:
                db_path = self.rii_loader['db_path']
                rel_path = os.path.join('database','data', page['data'])
                full_file = os.path.join(db_path, rel_path)
                try:
                    mat = Material(file_path=full_file,
                                       spectrum_type='wavelength',
                                       unit='micrometer')
                except OSError as exc:
                    warnings.warn("file {} ".format(full_file) +
                                  "could not be opened, skipping")
                    continue
                fullname = self.rii_loader['current_full_name']
                content_dict = {"Alias":"",
                                "Name":self.rii_loader['current_book'],
                                "FullName":fullname,
                                "Author":page['PAGE'],
                                "Path":os.path.normpath(rel_path),
                                "Module":"RefractiveIndexInfo"}
                content_dict['SpectrumType'] = 'wavelength'
                content_dict['Unit'] = 'micrometer'
                valid_range = mat.get_maximum_valid_range()
                content_dict['SpectrumLowerBound'] = valid_range[0]
                content_dict['SpectrumUpperBound'] = valid_range[1]
                content_dict['Reference'] = mat.meta_data['Reference']
                content_dict['Comment'] = mat.meta_data['Comment']
                try:
                    ref_spectrum = self.reference_spectrum
                    ref_index = mat.get_nk_data(ref_spectrum)
                except ValueError:
                    ref_index = np.nan + 1j* np.nan

                content_dict['N_Reference'] = np.real(ref_index)
                content_dict['K_Reference'] = np.imag(ref_index)
                self.rii_loader['database_list'].append(content_dict)

    def read_filmetrics_db(self, db_path):
        """read the file structure provided my filmetrics.com"""
        return self._read_text_db(db_path, "Filmetrics")

    def read_user_data_db(self, db_path):
        """read user data files"""
        return self._read_text_db(db_path, "UserData")

    def _read_text_db(self, db_path, database_name):
        """internal function used to load all txt, csv or yml
        files in a folder"""
        onlyfiles = [f for f in os.listdir(db_path)
                     if os.path.isfile(os.path.join(db_path, f))]

        dframe = pd.DataFrame(columns=Catalogue.META_DATA.keys())
        database_list = []
        for filename in onlyfiles:
            [name, ext] = os.path.splitext(filename)
            allowed_ext = {'.txt', '.csv', '.yml'}
            if ext not in allowed_ext:
                continue

            mat = Material(file_path=os.path.join(db_path, filename),
                               spectrum_type='wavelength',
                               unit='nanometer')

            content_dict = {}
            content_dict['Alias'] = ""
            content_dict['Name'] = name
            content_dict['FullName'] = mat.meta_data['FullName']
            content_dict['Author'] = mat.meta_data['Author']
            content_dict['Module'] = database_name
            content_dict['Comment'] = mat.meta_data['Comment']
            content_dict['Reference'] = mat.meta_data['Reference']
            content_dict['SpectrumType'] = "wavelength"
            content_dict['Unit'] = 'nanometer'
            valid_range = mat.get_maximum_valid_range()
            content_dict['SpectrumLowerBound'] = valid_range[0]
            content_dict['SpectrumUpperBound'] = valid_range[1]
            content_dict['Path'] = os.path.normpath(filename)
            try:
                ref_index = mat.get_nk_data(self.reference_spectrum)
            except ValueError as valE:
                ref_index = float('nan') + 1j* float('nan')

            content_dict['N_Reference'] = np.real(ref_index)
            content_dict['K_Reference'] = np.imag(ref_index)
            database_list.append(content_dict)
        dframe = pd.DataFrame(database_list,
                              columns=Catalogue.META_DATA.keys())
        return dframe




if __name__ == "__main__":
    rebuild_catalogue()



import numpy as np
import sys
import warnings
from dispersion.spectrum import Spectrum
from dispersion.spectral_data import *
from dispersion.io import Writer, Reader
from dispersion.material import Material
from dispersion.catalogue import Catalogue, rebuild_catalogue
from dispersion.config import get_config, default_config

__version__ = "1.0.4"
__all__ = ["Spectrum", "Material", "Catalogue", "get_config",
           "Constant", "Interpolation", "Extrapolation",
           "Sellmeier", "Sellmeier2", "Polynomial",
           "RefractiveIndexInfo", "Cauchy", "Gases",
           "Herzberger", "Retro", "Exotic", "Drude",
           "DrudeLorentz", "rebuild_catalogue", "Writer", "Reader"]

try:
    config = get_config()
except ValueError:
    config = default_config()

try:
    from ruamel.yaml import YAML
    from ruamel.yaml import scalarstring
    from ruamel.yaml.comments import CommentedMap
except ModuleNotFoundError as exc:
    if config['WarnMissingPackages']:
        warnings.warn("preferred yaml package ruamel.yaml not installed, " +
                      "falling back to PyYAML, writing yaml files may give " +
                      "inconsistent round trip results")



"""spectrum is an array of data which defines a physical spectrum

A Spectrum is a numpy array with an associated type which describes
the physical quantity (e.g. wavelength, energy, etc.) and a unit
for specifiying the SI unit of the spectrum (e.g. meter, electronVolt)
"""
import numpy as np
import scipy.constants as constants

def safe_inverse(values):
    '''
    returns the inverse of a scalar or array while
    converting values of 0 to inf and values of inf to 0
    '''
    if isinstance(values,np.ndarray):
        inverse = safe_inverse_array(values)
    else:
        inverse = safe_inverse_scalar(values)
    return inverse

def safe_inverse_array(values):
    '''
    takes the inverse of an array while converting values of 0 to inf
    and values of inf to 0
    '''
    inverse = np.array(values)
    zero_ind = np.isclose(values,0.0)
    inf_ind = np.isclose(values,np.inf)
    safe_inv_ind = ~ (zero_ind | inf_ind)
    inverse[zero_ind] = np.inf
    inverse[inf_ind] = 0.0
    inverse[safe_inv_ind] = 1.0/values[safe_inv_ind]
    return inverse

def safe_inverse_scalar(value):
    '''
    takes the inverse of a scalar while converting values of 0 to inf
    and values of inf to 0
    '''
    if value == 0.0:
        return np.inf
    elif value == np.inf:
        return 0.0
    else:
        return 1.0/value

def frequency_to_standard(unit, values):
    """converts frequency values to standardised representation"""
    if not unit == 'hz':
        raise NotImplementedError("support for frequency units other " +
                                  "than Hz not implemented")
    inverse = safe_inverse(values)
    converted_values = constants.c*inverse
    return converted_values


def energy_to_standard(unit, values):
    """converts energy values to standardised representation"""
    if not unit == 'ev':
        raise NotImplementedError("support for energy units other " +
                                  "than eV not implemented")
    inverse = safe_inverse(values)
    converted_values = (constants.h*constants.c/(constants.e))*inverse
    return converted_values


def ang_freq_to_standard(unit, values):
    """converts angular frequency values to standardised representation"""
    if not unit == '1/s':
        raise NotImplementedError("support for angular frequency " +
                                  "units other than 1/s " +
                                  "not implemented")
    inverse = safe_inverse(values)
    converted_values = 2*np.pi*constants.c*inverse
    return converted_values


def wavenumber_to_standard(unit, values):
    """converts wavenumber values to standardised representation"""
    if not unit == '1/cm':
        raise NotImplementedError("support for wavenumber units " +
                                  "other than 1/cm not implemented")
    inverse = safe_inverse(values)
    converted_values = 2*np.pi*inverse*1e-2
    return converted_values


def wavelength_to_standard(unit, values):
    """converts wavelength values to standardised representation"""
    if unit == 'nm':
        converted_values = values*1e-9
    elif unit == 'cm':
       converted_values = values*1e-2
    elif unit == 'um':
        converted_values = values*1e-6
    elif unit =='a':
        converted_values = values*1e-10
    else:
        converted_values = values
    return converted_values

def to_wavelength(unit, values):
    """converts values from standard to wavelength representation"""
    if unit == 'm':
        converted_values = values
    elif unit == 'cm':
        converted_values = np.round(values*1e2, decimals=5)
    elif unit == 'nm':
        converted_values = np.round(values*1e9, decimals=12)
    elif unit == 'um':
        converted_values = np.round(values*1e6, decimals=9)
    elif unit == 'a':
        converted_values = np.round(values*1e10, decimals=13)
    return converted_values

def to_frequency(unit, values):
    """converts values from standard to frequency representation"""
    if not unit == 'hz':
        raise NotImplementedError("support for frequency units " +
                                  "other than Hz not implemented")
    inverse = safe_inverse(values)
    return constants.c * inverse

def to_energy(unit, values):
    """converts values from standard to energy representation"""
    if not unit == 'ev':
        raise NotImplementedError("support for energy units " +
                                  " other than eV not implemented")
    inverse = safe_inverse(values)
    return np.round((constants.h*constants.c/(constants.e))*inverse, decimals=9)

def to_ang_freq(unit, values):
    """converts values from standard to angular frequency representation"""
    if not unit == '1/s':
        raise NotImplementedError("support for angular frequency units " +
                                  "other than 1/s not implemented")
    inverse = safe_inverse(values)
    return 2*np.pi*constants.c*inverse

def to_wavenumber(unit, values):
    """converts values from standard to wavenumber representation"""
    if not unit == '1/cm':
        raise NotImplementedError("support for wavenumber units" +
                                  " other than 1/cm not implemented")
    inverse = safe_inverse(values*1e-2)
    return 2*np.pi*inverse

class Spectrum(object):
    '''
    Class for converting between wavelength/frequency/energy
    when defining spectral quantities
    '''

    SPECTRUM_TYPES = {'wavelength': {'m', 'cm', 'um', 'nm', 'a'},
                      'frequency': {'hz'},
                      'energy': {'ev'},
                      'angularfrequency': {'1/s'},
                      'wavenumber': {'1/cm'}}

    UNIT_TYPES = {'m': {'meter', 'metre', 'm'},
                  'cm': {'centimeter', 'centimetre', 'cm'},
                  'um': {'micrometer', 'micrometre', 'um'},
                  'nm': {'nanometer', 'nanometre', 'nm'},
                  'hz': {'hertz', 'hz'},
                  'ev': {'electronvolt', 'ev'},
                  '1/s': {'1/s', 'rad/s'},
                  '1/cm': {'1/cm'},
                  'a': {'a', 'angstrom'}}

    DISPLAY_NAMES = {'m': 'm',
                     'um': u'μm',
                     'nm': 'nm',
                     'hz': 'Hz',
                     'ev': 'eV',
                     '1/s': '1/s',
                     '1/cm': '1/cm',
                     'a':u'Å'}

    def __init__(self, values, spectrum_type='wavelength', unit='m'):
        self.spectrum_type = spectrum_type.lower()
        if self.spectrum_type not in Spectrum.SPECTRUM_TYPES:
            spectral_types = list(Spectrum.SPECTRUM_TYPES.keys())
            raise ValueError("spectrum type [{}] ".format(spectrum_type) +
                             "not valid, valid types are " +
                             "{}".format(spectral_types))
        self.unit = self.standardise_unit(self.spectrum_type, unit.lower())
        if isinstance(values, (list, tuple)):
            values = np.array(values)
        if not ( isinstance(values, np.ndarray) or
                 isinstance(values, float)):
            raise ValueError("values must be array like or float")

        self.values = values
        self.standard_rep = None  # values in standardised representation
        self.standardise_values(values)

    @staticmethod
    def standardise_unit(spectrum_type, unit):
        """takes a spectrum_type and unit string and returns them in a
        standardised form"""
        valid = False
        unit_type = ""
        for unit_type, aliases in Spectrum.UNIT_TYPES.items():
            if (unit in aliases and
                    unit_type in Spectrum.SPECTRUM_TYPES[spectrum_type]):
                valid = True
                break
        if valid is False:
            s_types = Spectrum.SPECTRUM_TYPES
            raise ValueError("unit [{}] invalid ".format(unit) +
                             "valid units for type {} ".format(spectrum_type) +
                             "are {}".format(s_types[spectrum_type]))
        return unit_type

    def get_type_unit_string(self):
        """formats the unit and spectrum_type for display as an axis label"""
        unit_str = self.display_string(self.unit)
        type_str = self.spectrum_type
        return "{} ({})".format(type_str, unit_str)

    def contains(self, other):
        """detects if the values of another Spectrum object are
        completely contained within the range of the values defined
        in this Spectrum object"""
        other_values = other.convert_to(self.spectrum_type, self.unit)
        other_min = np.min(other_values)
        other_max = np.max(other_values)
        self_min = np.min(self.values)
        self_max = np.max(self.values)
        out_of_range = None
        if other_min < self_min:
            out_of_range = other_min
        elif other_max > self_max:
            out_of_range = other_max
        if out_of_range is not None:
            dsu = self.display_string(self.unit)  # display string unit
            raise ValueError("evaluation requested for " +
                             "{} {} {}, ".format(self.spectrum_type,
                                                 out_of_range,
                                                 dsu) +
                             "but valid range is {} {} < {} < {} {}".format(
                                 self_min, dsu, self.spectrum_type,
                                 self_max, dsu))

        return True

    @staticmethod
    def check_type(type_to_check, is_type):
        """test if two spectrum_type strings are equal taking aliases
        into account"""
        test_str = type_to_check.lower()
        if test_str not in Spectrum.SPECTRUM_TYPES:
            raise ValueError("spectrum type " +
                             "{} not valid, valid types are: {}".format(
                                 test_str, Spectrum.SPECTRUM_TYPES.keys()))
        return test_str == is_type

    @staticmethod
    def check_unit(unit, is_unit):
        """test if two unit strings are equal taking aliases
        into account"""
        test_str = unit.lower()
        return test_str == is_unit

    @staticmethod
    def display_string(unit):
        """gets the SI standard form of a unit string for printing"""
        return Spectrum.DISPLAY_NAMES[unit]

    def standardise_values(self, values):
        """check if we need to convert the value in order to obtain
        the wavelength/meter representation"""
        if not (self.spectrum_type == 'wavelength' and
                self.unit == 'm'):
            self.standard_rep = self.convert_from(self.spectrum_type,
                                                  self.unit, values)
        else:
            self.standard_rep = values

    @staticmethod
    def convert_from(spectrum_type, unit, values):
        """
        converts units from given type into wavelength/meter representation
        """
        if spectrum_type not in Spectrum.SPECTRUM_TYPES:
            raise ValueError("invalid spectrum type {}".format(spectrum_type))

        convert_methods = {'frequency': frequency_to_standard,
                           'energy': energy_to_standard,
                           'angularfrequency': ang_freq_to_standard,
                           'wavenumber': wavenumber_to_standard,
                           'wavelength': wavelength_to_standard}

        converted_values = convert_methods[spectrum_type](unit, values)
        return converted_values

    def convert_to(self, spectrum_type, unit, in_place=False):
        """
        converts units from wavelength/meter representation to given type/unit
        """
        spectrum_type = spectrum_type.lower()
        if spectrum_type not in Spectrum.SPECTRUM_TYPES:
            raise ValueError("invalid spectrum type {}".format(spectrum_type))

        unit = Spectrum.standardise_unit(spectrum_type, unit)
        values = self.standard_rep

        convert_methods = {'frequency': to_frequency,
                           'energy': to_energy,
                           'angularfrequency': to_ang_freq,
                           'wavenumber': to_wavenumber,
                           'wavelength': to_wavelength}

        converted_values = convert_methods[spectrum_type](unit, values)
        if in_place:
            self.values = converted_values
            self.spectrum_type = spectrum_type
            self.unit = unit
            return None
        else:
            return converted_values



"""provides functions for reading and writing the configuration file."""
#import sys
import os
from warnings import warn
from dispersion.io import (read_yaml_file, read_yaml_string,
                           write_yaml_file, valid_file_name)
if os.name == 'nt':
    PLATFORM = "Windows"
elif os.name == 'posix':
    PLATFORM = 'Linux'
else:
    raise OSError("curret os type could not be determined."+
                  " Configuration file location unknown.")


def validate_config(config):
    """
    validates the data types of the configuration

    Parameters
    config: dict
        dictionary containing all configuration
    """
    check_type(config['Path'],str)
    if not os.path.isdir(config['Path']):
        raise IOError("directory path for database file system is invalid:" +
                      " <{}>".format(config['Path']))
    check_type(config['File'],str)
    if not valid_file_name(config['File']):
        raise IOError("file path for database file is invalid:" +
                      " <{}>".format(config['File']))
    check_type(config['Interactive'],bool)
    check_type(config['WarnMissingPackages'],bool)
    assert "Modules" in config
    for value in config['Modules'].values():
        check_type(value,bool)
    assert "ReferenceSpectrum" in config
    ref_spec = config['ReferenceSpectrum']
    check_type(ref_spec['Value'],float)
    check_type(ref_spec['SpectrumType'],str)
    check_type(ref_spec['Unit'],str)

def check_type(value, val_type):
    """
    checks if value isinstance of val_type, if not raise exception
    """
    if not isinstance(value,val_type):
        raise ValueError("config file data {}".format(value) +
                         "must be of type {}".format(val_type) +
                         ", not {}".format(type(value)))


def default_config():
    """provides default values for the configuration.

    Returns
    -------
    dict or OrderedDict
        the tree of configuration data
    """
    yaml_str = """\
    Path: {}
    File: catalogue.csv
    Interactive: false #for jupyter interfactive editing
    WarnMissingPackages: False # raise warnings if non-essential packages are missing
    Modules: # which databases to include
      UserData: true
      RefractiveIndexInfo: true
    ReferenceSpectrum: # evaluate n and k at given spectral value
      Value: 632.8
      SpectrumType: wavelength
      Unit: nanometer
    """.format(_get_user_config_dir())
    config = read_yaml_string(yaml_str)
    return config

def _get_user_config_dir():
    if PLATFORM == 'Windows':
        user_dir = os.environ["LOCALAPPDATA"]
        config_dir = os.path.join(user_dir, "dispersion")
    elif PLATFORM == 'Linux':
        home_dir = os.environ["HOME"]
        config_dir = os.path.join(home_dir, ".config",
                                  "dispersion")
    if not os.path.exists(config_dir):
        os.mkdir(config_dir)
    return config_dir

def _get_package_dir():
    dir_path = os.path.dirname(os.path.realpath(__file__))
    return dir_path

def _get_config_dir():
    user_dir = _get_user_config_dir()
    file_path = os.path.join(user_dir, 'config.yaml')
    if os.path.isfile(file_path):
        return user_dir
    pkg_dir = _get_package_dir()
    file_path = os.path.join(pkg_dir, 'config.yaml')
    if os.path.isfile(file_path):
        return pkg_dir
    raise ValueError("could not locate configuration or package directory, "+
                     "unable to load configuration file.")


def write_config(config, dir_path=None):
    """write the configuration data to file.

    Parameters
    ----------
    config: dict or OrderedDict
        the configuration data

    Warnings
    --------
    No check is made if the config is valid or complete.
    """
    if dir_path is None:
        dir_path = _get_config_dir()
    file_path = os.path.join(dir_path, 'config.yaml')
    write_yaml_file(file_path, config)

def read_config():
    """read the configuration data from file.

    Returns
    -------
    dict or OrderedDict
        the configuration data
    """
    dir_path = _get_config_dir()
    file_path = os.path.join(dir_path, 'config.yaml')
    config = read_yaml_file(file_path)
    return config

def get_config():
    """get the configuration data.

    attempt to read the configuration from file, if not found then return
    default values.

    Returns
    -------
    dict or OrderedDict
        the configuration data
    """
    try:
        config = read_config()
    except OSError as exc:
        warn("No configuration file found, generating default config.")
        config = default_config()
    return config

if __name__ == "__main__":
    print(get_config())



#!/usr/bin/env python
"""
setup the disperion database file structure and configuration file
"""

from dispersion import rebuild_catalogue

def main():
    rebuild_catalogue()

if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
setup the disperion database file structure and configuration file
"""
import os
import numpy as np
from dispersion import Material, Writer, Interpolation, Catalogue
from dispersion.config import default_config, write_config, _get_user_config_dir
from dispersion.io import valid_file_name

def get_root_dir(conf):
    """
    get the root dir from the user
    """
    question = ("path for root directory of the catalogue file" +
                " system [default: {}]> ")
    default = conf['Path']
    validator = os.path.isabs
    data_name = "root directory"
    return ask_and_confirm(question, default, validator, data_name)

def get_catalogue_name(conf):
    """
    get the catalogue file name from the user
    """
    question = ("name of the catalogue file" +
                " [default: {}]> ")
    default = conf['File']
    validator = valid_file_name
    data_name = "catalogue file name"
    return ask_and_confirm(question, default, validator, data_name)

def ask_and_confirm(question, default, validator, data_name, confirm=True):
    """
    Returns
    -------
    user_input: str
        the data from the user
    confirmed_input: bool
        true if the input was confirmed by the user

    Parameters
    ----------
    question: str
        the question to prompt the user input
    default: str
        the default value of this value
    validator: function
        function to validate the input
    data_name: str
        name of the data that is being input
    """
    user_input = ask(question, default, validator)
    confirmation_question = ("confirm {} as ".format(data_name) +
                             "{}? [y/n]> ".format(user_input))
    return [user_input, get_confirmation(confirmation_question)]


def ask(question, default, validator):
    """
    ask for user input with default value and then validate
    """
    valid_input = False
    while not valid_input:
        user_input = input(question.format(default))
        if user_input == "":
            user_input = default
        if validator(user_input):
            valid_input = True
        else:
            print("input is not valid")
    return user_input

def get_confirmation(question):
    """
    get a yes/no answer to a question
    """
    confirmed_input = False
    while not confirmed_input:
        confirmation1 = input(question)
        if confirmation1 in {'y', 'yes'}:
            confirmed_input = True
        elif confirmation1 in {'n', 'no'}:
            confirmed_input = False
            break
        else:
            print("input invalid")
    return confirmed_input


def install_modules(conf):
    """
    make a subfolder for each module and ask to download files
    """

    install_funcs = {"UserData":install_userdata,
                     "RefractiveIndexInfo":install_rii}

    for module in conf['Modules']:
        if module == "UserData":
            install = True
        else:
            question = "install module {}? [y/n]> ".format(module)
            install = get_confirmation(question)

        conf['Modules'][module] = install
        if install:
            module_dir = os.path.join(conf['Path'], module)
            if not os.path.isdir(module_dir):
                os.mkdir(module_dir)
            install_funcs[module](module_dir)
    return conf

def install_userdata(module_dir):
    make_example_txt(module_dir)
    make_example_yaml(module_dir)

def make_example_txt(dir_path):
    test_data = np.array([[400., 1.7, 0.1],
                          [500., 1.6, 0.05],
                          [600., 1.5, 0.0],
                          [700., 1.4, 0.0]])
    mat = Material(tabulated_nk=test_data,
                   spectrum_type='wavelength', unit='nanometer')
    mat.meta_data['Reference'] = "Literature reference to the data"
    mat.meta_data['Comment'] = "Any additional information goes here"
    mat.meta_data['Name'] = "Short name of the material"
    mat.meta_data['FullName'] = "Full name of the material"
    mat.meta_data['Author'] = "The author of this data file"
    mat.meta_data['MetaComment'] = " This is a multiline meta-comment\n" + \
                                   " which provides information not\n" + \
                                   " in metadata"
    filepath = os.path.join(dir_path, "example_file.txt")
    write = Writer(filepath, mat)
    write.write_file()

def make_example_yaml(dir_path):
    model_params = {'name': 'Sellmeier',
                    'specrtrum_type':'wavelength',
                    'unit':'micrometer',
                    'valid_range':np.array([0.350, 2.0]),
                    'parameters': np.array([0, 1.0, 0.05,
                                            2.0, 0.1,
                                            10., 25.])}
    mat = Material(model_kw=model_params, spectrum_type='wavelength', unit='micrometer')
    mat.meta_data['Reference'] = "Literature reference to the data"
    mat.meta_data['Comment'] = "Any additional information goes here"
    mat.meta_data['Name'] = "Short name of the material"
    mat.meta_data['FullName'] = "Full name of the material"
    mat.meta_data['Author'] = "The author of this data file"
    mat.meta_data['MetaComment'] = " This is a multiline meta-comment\n" + \
                                   " which provides information not\n" + \
                                   " in metadata"
    k_data = np.array([[400., 0.1],
                       [500., 0.05],
                       [600., 0.0],
                       [700., 0.0]])
    interp = Interpolation(k_data, unit='nm')
    mat.data['imag'] = interp
    filepath = os.path.join(dir_path, "example_file2.yml")
    write = Writer(filepath, mat)
    write.write_file()

def install_rii(module_dir, ask_confirm=True):
    """
    download the refractive index info database from github
    """
    question = ("download the refractive index info database from github?" +
                " (required python package <GitPython>)" +
                " [y/n]> ")
    if ask_confirm:
        install = get_confirmation(question)
    else:
        install = True
    if install:
        from git import Repo
        git_url = "https://github.com/polyanskiy/refractiveindex.info-database.git"
        Repo.clone_from(git_url, module_dir)

def maybe_rebuild_catalogue(conf):
    question = "rebuild catalogue? [y/n]> "
    rebuild = get_confirmation(question)
    if rebuild:
        cat = Catalogue(config=conf, rebuild= 'All')
        cat.save_to_file()

def main():
    conf = default_config()
    print("This script will provide a default configuration for the \n"+
          "dispersion package")
    confirmed_valid_path = False
    while not confirmed_valid_path:
        [path, confirmed_valid_path] = get_root_dir(conf)
    conf['Path'] = path
    #print("Path will be se to: {}".format(path))
    confirmed_db_nane = False
    while not confirmed_db_nane:
        [name, confirmed_db_nane] = get_catalogue_name(conf)
    conf['File'] = name

    #print("Filename will be set to {}".format(name))
    conf = install_modules(conf)
    user_dir = _get_user_config_dir()
    write_config(conf, dir_path=user_dir)

    maybe_rebuild_catalogue(conf)





if __name__ == "__main__":
    main()



"""provides functions for reading and writing files.

implements a set of functions for reading and writing yaml files. Also
implements a reader and writer object for interfacing with material data files.

Functions
---------
validate_table
    checks that tabulated data can be interpolated
fix_table
    removes rows of table which stop interpolation from happening
_str_table_to_numeric
    convert a string table to a numpy array
read_yaml_file
    convert a file with yaml format to dict
read_yaml_string
    convert a string with yaml format to dict
write_yaml_file
    write dict to a yaml format file
print_yaml_string
    print a dict in yaml format to stdout
prepend_text_to_file
    add text to start of a file
valid_file_name
    check if file name is valid for this operating system

Classes
-------
Reader
    reads refractive index data from file
Writer
    writes refractive index data to file
"""
import os
import sys
import codecs
import re
import tempfile
from collections import OrderedDict
import numpy as np
USE_RUAMEL = True
try:
    from ruamel.yaml import YAML
    from ruamel.yaml import scalarstring
    from ruamel.yaml.comments import CommentedMap
except ModuleNotFoundError as exc:
    USE_RUAMEL = False
    import yaml
#from dispersion.material import _str_table_to_numeric
#USE_RUAMEL = False
#import yaml

def multi_key(dict_obj, key_set):
    for key in key_set:
        if key in dict_obj:
            return dict_obj[key]
    else:
        raise KeyError("None of the keys {}".format(key_set) +
                       " were present")

def _numeric_to_string_table(table):
    """
    returns a string representation of a numpy array
    """
    string_array = np.array2string(table)
    no_brackets = re.sub(r'[\[\]]', '', string_array)
    no_indent = re.sub(r'\n ', r"\n", no_brackets)
    no_indent = no_indent.rstrip().lstrip()
    return no_indent

def _str_table_to_numeric(table):
    '''
    takes tabulated data in string form
    and converts to a numpy array
    '''

    if isinstance(table, np.ndarray):
        numeric_table = table
    elif isinstance(table, str):
        #table is a str
        numeric_table = []
        for row in table.split('\n'):
            if row.isspace() or row == "":
                break
            numeric_col = []
            for col in row.split():
                numeric_col.append(float(col))
            numeric_table.append(numeric_col)
        numeric_table = np.array(numeric_table)

    else:
        raise TypeError("table of type " +
                        "{} cannot be parsed".format(type(table)))
    if validate_table(numeric_table) is False:
        numeric_table = fix_table(numeric_table)
    return numeric_table



def validate_table(tabulated_data):
    '''
    check that spectral part (first column) is
    monotonically increasing to be able to interpolate
    '''
    return np.all(tabulated_data[1:, 0] > tabulated_data[:-1, 0])


def fix_table(tabulated_data):
    '''
    throw out rows which break strict monotonicity
    '''
    n_cols = tabulated_data.shape[1]
    new_rows = [tabulated_data[0, :]]
    last_valid = tabulated_data[0, 0]
    for row in range(1, tabulated_data.shape[0]):
        if not tabulated_data[row, 0] > last_valid:
            continue
        else:
            new_rows.append(tabulated_data[row, :])
            last_valid = tabulated_data[row, 0]
    return np.array(new_rows).reshape(-1, n_cols)

def read_yaml_file(file_path):
    """opens yaml file and returns contents as a dict like.

    Parameters
    ----------
    file_path: str
        path to the yaml file

    Returns
    -------
    dict or OrderedDict
        the data from the yaml file

    Notes
    -----
    If USE_RUAMEL is true an OrderedDict is returned, otherwise a dict is
    returned.
    """
    with open(file_path, 'r', encoding="utf-8") as fpt:
        if USE_RUAMEL:
            yaml_obj = YAML()
            yaml_data = yaml_obj.load(fpt)
        else:
            try:
                yaml_data = yaml.load(fpt, Loader=yaml.FullLoader)
            except:
                yaml_data = yaml.safe_load(fpt)
    return yaml_data

def read_yaml_string(string_data):
    """converts a string in yaml format to a dict like.

    Parameters
    ----------
    string_data: str
        string in the yaml format

    Returns
    -------
    dict or OrderedDict
        the data from the yaml file

    Warnings
    --------
    No attempt is made to check if string_data is in correct yaml format.

    Notes
    -----
    If USE_RUAMEL is true an OrderedDict is returned, otherwise a dict is
    returned.
    """
    if USE_RUAMEL:
        yaml_obj = YAML()
        yaml_data = yaml_obj.load(string_data)
    else:
        try:
            yaml_data = yaml.load(string_data, Loader=yaml.FullLoader)
        except:
            yaml_data = yaml.safe_load(string_data)
    return yaml_data

def write_yaml_file(file_path, dict_like):
    """write a dict_like object to a file using the given file path.

    Parameters
    ----------
    file_path: str
        path with file name and extension
    dict_like: dict or OrderedDict
        the data to be written to file
    """
    with open(file_path, 'w', encoding='utf8') as fpt:
        if USE_RUAMEL:
            yaml_obj = YAML()
            yaml_obj.indent(mapping=4, sequence=4, offset=2)
            scalarstring.walk_tree(dict_like)
            yaml_obj.dump(dict_like, fpt)
        else:
            yaml.dump(dict_like, fpt)

def print_yaml_string(dict_like):
    """print a dict_like object to stdout in yaml format.

    Parameters
    ----------
    dict_like: dict or OrderedDict
        the data to be written to printed
    """
    if USE_RUAMEL:
        yaml_obj = YAML()
        yaml_obj.dump(dict_like, sys.stdout)
    else:
        print(yaml.dump(dict_like))

def prepend_text_to_file(file_path, text, extra_line=False):
    """
    read information from a file and rewrite with text prepended
    """
    with open(file_path, 'r', encoding='utf8') as fpt:
        current_text = fpt.read()

    with open(file_path, 'w', encoding='utf8') as fpt:
        text = text.rstrip('\n\r')
        text = text.replace("\n", "\n#")
        text = "#" + text
        text += "\n"
        if extra_line:
            text += "\n"
        fpt.write(text)
        fpt.write(current_text)

def valid_file_name(filename):
    """
    test if filename is valid

    create a file with the filename in a temporary directory and delete the
    directory afterwards.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        file_path = os.path.join(temp_dir, filename)
        try:
            open(file_path, 'r')
            return True
        except IOError:
            try:
                open(file_path, 'w')
                return True
            except IOError:
                return False


class Reader():
    """interface for reading refractive index data from file.

    Attributes
    ----------
    file_path: str
        path to the file to be read
    extension: str
        file type to read
    default_file_dict: dict
        default values to use

    Methods
    -------
    read_file
        read the associated file
    """

    FILE_META_DATA_KEYS = {"Comment", "Reference", "Author",
                           "Name", "FullName", "MetaComment"}

    DATASET_META_DATA_KEYS = {'ValidRange', 'DataType',
                              'SpectrumType', 'Unit',
                              'Yields'}

    def __init__(self, file_path):
        self.file_path = file_path
        fname, extension = os.path.splitext(file_path)
        self.extension = extension
        self.default_file_dict = self._create_default_file_dict()

    def _create_default_file_dict(self):
        """default values for the material data."""
        file_dict = {'MetaData': {}}
        for mdk in Reader.FILE_META_DATA_KEYS:
            file_dict['MetaData'][mdk] = ""
        file_dict['Datasets'] = [Reader._create_default_data_dict()]
        file_dict['MetaData']['Specification'] = {}
        #file_dict['MetaComment'] = ""
        file_dict['FilePath'] = self.file_path
        return file_dict

    @staticmethod
    def _create_default_data_dict():
        """default values for a data set."""
        dataset_dict = {}
        for mdk in Reader.DATASET_META_DATA_KEYS:
            dataset_dict[mdk] = ""
        dataset_dict['Data'] = []
        return dataset_dict

    def read_file(self):
        """reads the material data file from file.

        Returns
        -------
        dict
            the data from the material file
        """
        txt_types = {'.txt', '.csv', '.nk'}
        if self.extension in txt_types:
            return self._read_text_file()
        elif self.extension == '.yml':
            return self._read_yaml_mat_file()
        else:
            raise ValueError("extension " +
                             "{} not supported".format(self.extension) +
                             ", supported extensions are (.yml|.csv|.txt)")

    def _read_text_data(self):
        """read data stored in a .txt, .csv or .nk file."""
        fname, ext = os.path.splitext(self.file_path)
        try:
            if ext == '.txt':
                try:
                    data = np.loadtxt(self.file_path, encoding='utf-8')
                except TypeError:
                    with open(self.file_path, encoding = 'utf-8') as fpt:
                        data = np.loadtxt(fpt)
            elif ext == '.csv':
                try:
                    data = np.loadtxt(self.file_path, encoding='utf-8',
                                      delimiter=',')
                except TypeError:
                    with open(self.file_path, encoding = 'utf-8') as fpt:
                        data = np.loadtxt(fpt,delimiter=',')
            elif ext == '.nk':
                try:
                    data = np.loadtxt(self.file_path, encoding='utf-8',
                                      comments=';')
                except TypeError:
                    with open(self.file_path, encoding = 'utf-8') as fpt:
                        data = np.loadtxt(fpt, comments=';')
        except IOError as exc:
            raise exc
        data_dict = self._create_default_data_dict()
        data_dict['Data'] = data
        return data_dict

    def _read_text_comment(self):
        '''
        returns text from contiguous lines in file beginning with #.
        '''
        comment = []
        with codecs.open(self.file_path, 'r', 'utf-8') as fpt:
            for line in fpt:
                if line[0] == "#" or line[0] == ';':
                    comment.append(line[1:].rstrip("\n\r"))
                else:
                    return comment
        return comment

    def _read_text_file(self):
        """
        text files (.txt,.csv,.nk) may only contain tabulated nk data
        plus metadata. Metadata should be at the beginning of the
        file, one item per line. The line must start with '#' or ';' to
        denote metadata. Metadata is written in the key:value
        structure.
        """

        dataset = self._read_text_data()
        comment = self._read_text_comment()
        file_dict = dict(self.default_file_dict)
        file_dict['Datasets'][0] = dataset
        multi_line_comment = ""
        for line in comment:
            kwd_arg = line.split(":")
            kwd = kwd_arg[0].upper().lstrip()
            if len(kwd_arg) == 1:
                multi_line_comment += line + "\n"
            else:
                arg = ":".join(kwd_arg[1:]).rstrip("\n\r").lstrip()
                #arg = kwd_arg[1].rstrip("\n\r").lstrip()
                valid_key_val = False
                for key in Reader.FILE_META_DATA_KEYS:
                    if kwd.startswith(key.upper()):
                        if key == 'Specification':
                            file_dict['MetaData'][key] = str(arg)
                        else:
                            file_dict['MetaData'][key] = arg
                        valid_key_val = True
                        break
                if valid_key_val is False:
                    for key in Reader.DATASET_META_DATA_KEYS:
                        if kwd.startswith(key.upper()):
                            file_dict['Datasets'][0][key] = arg
                            valid_key_val = True
                            break
                #if valid_key_val is False:
                #    KeyError("keyword " +
                #             "[{}] in comment header invalid".format(kwd))
                #else:
                if valid_key_val is False:
                    multi_line_comment += line + "\n"

        if multi_line_comment != "":
            file_dict['MetaData']['MetaComment'] = multi_line_comment
        dataset = file_dict['Datasets'][0]

        if dataset['DataType'] == "":
            if dataset['Data'].shape[1] == 3:
                dataset['DataType'] = 'tabulated nk'
            elif dataset['Data'].shape[1] == 2:
                dataset['DataType'] = 'tabulated n'
        return file_dict

    def _read_yaml_mat_file(self):
        '''
        The refractiveindex.info database format
        '''
        #yaml_stream = open(self.file_path, 'r', encoding="utf-8")
        #dir_path = os.path.dirname(os.path.realpath(__file__))
        yaml_data = read_yaml_file(self.file_path)

        file_dict = dict(self.default_file_dict)
        file_dict = self._process_mat_dict(file_dict, yaml_data)
        mcomment = "\n".join(self._read_text_comment())
        mcomment = mcomment[:-1]
        file_dict['MetaData']['MetaComment'] = mcomment
        return file_dict

    def _process_mat_dict(self, file_dict, yaml_dict):
        """
        copies raw data from a yaml file to the format used in this package.

        Parameters
        ----------
        file_dict: dict
            dict with the correct field names for creating a materialdata object
        yaml_dict: dict
            dict with the raw data describing a material

        Returns
        -------
        dict
            the updated file_dict
        """
        for kwd in yaml_dict:
            kwd = kwd.upper()
            arg = yaml_dict[kwd]
            valid = False
            if kwd == 'DATA':
                valid = True
                file_dict['Datasets'] = self._process_mat_data_dict(arg)
            elif kwd == 'SPECS':
                valid = True
                file_dict['MetaData']['Specification'] = arg
            if valid is False:
                for key in Reader.FILE_META_DATA_KEYS:
                    if kwd.startswith(key.upper()):
                        file_dict['MetaData'][key] = arg
                        valid = True
                        break
            if valid is False:
                KeyError("keyword [{}] in file invalid".format(kwd))
        return file_dict

    def _process_mat_data_dict(self, mat_data):
        """
        creates a list of dicts to generate spectral data sets

        Parameters
        ----------
        mat_data: list of dicts
            data for creating spectral data sets

        Returns
        -------
        list of dicts
            formated data for use in this package
        """
        aliases = {'ValidRange': {'validRange', 'range',
                                  'spectra_range', 'wavelength_range'},
                   'DataType':'type'}
        dataset_list = []
        for n_data_set, dataset in enumerate(mat_data):
            dataset_list.append(Reader._create_default_data_dict())
            data_dict = dataset_list[-1]
            try:
                data_type = dataset['type'].lstrip()
            except KeyError:
                data_type = dataset['DataType'].lstrip()
            data_dict['DataType'] = data_type
            if (data_type.startswith('formula') or
                    data_type.startswith('model')):
                try:
                    data_dict['Data'] = dataset['coefficients']
                except KeyError:
                    data_dict['Data'] = dataset['Parameters']
            elif data_type.startswith('tabulated'):
                try:
                    data_dict['Data'] = dataset['data']
                except KeyError:
                    data_dict['Data'] = dataset['Data']
            else:
                raise KeyError("data type <{}> invalid".format(data_type))

            for kwd in dataset:
                valid = False
                arg = dataset[kwd]
                kwd = kwd.upper()
                for key in Reader.DATASET_META_DATA_KEYS:
                    if valid:
                        break
                    if key in aliases.keys():
                        all_names = aliases[key]
                    else:
                        all_names = {key}
                    for alias in all_names:
                        if kwd == alias.upper():
                            data_dict[key] = arg
                            valid = True
                            break
                if valid is False:
                    KeyError("keyword <{}> in file invalid".format(kwd))

        return dataset_list


class Writer():
    """interface for writing refractive index data to file.

    Attributes
    ----------
    file_path: str
        path to the file to be written
    extension: str
        file type to write
    file_name: str
        name of the file
    material: MaterialData
        the material object to be written

    Methods
    -------
    write_file
        write the material data to file

    Notes
    -----
    if data_dict is not an OrderedDict, the order in which the data and
    metadata is written cannot be controlled.
    """

    KEY_ALIASES = {"Reference": "REFERENCES",
                   "Author": "AUTHOR",
                   "FullName": "FULLNAME",
                   "Name": "NAME",
                   "Comment": "COMMENTS",
                   "Specification": "SPECS",
                   "Datasets": "DATA"}

    RII_ALIASES = {"ValidRange": "wavelength_range",
                   "DataType": "type",
                   "Parameters": "coefficients",
                   "Data":"data",
                   "Model:Sellmeier":"formula 1",
                   "Model:Sellmeier2":"formula 2",
                   "Model:Polynomial":"formula 3",
                   "Model:RefractiveIndexInfo":"formula 4",
                   "Model:Cauchy":"formula 5",
                   "Model:Gases":"formula 6",
                   "Model:Herzberger":"formula 7",
                   "Model:Retro":"formula 8",
                   "Model:Exotic":"formula 9"}

    DATA_META_DATA_KEYS = ['Reference', 'Author',
                           'FullName', 'Name', 'Comment']

    DATASET_META_DATA_KEYS = ['ValidRange', 'DataType',
                              'SpectrumType', 'Unit']

    def __init__(self, file_path, material):
        self.file_path = file_path
        fname, extension = os.path.splitext(file_path)
        self.extension = extension
        self.file_name = fname
        self.file_dict = material.prepare_file_dict()
        self.ignore_constant = True
        self.use_rii_aliases = False


    def write_file(self, ignore_constant=True,
                   use_rii_aliases=False):
        """
        write the data to the file_path
        """
        self.ignore_constant = ignore_constant
        self.use_rii_aliases = use_rii_aliases
        meta_comment = self.file_dict['MetaData'].pop('MetaComment')
        #raise NotImplementedError("writing material files not yet implemented")
        txt_types = {'.txt', '.csv', '.nk'}
        if self.extension in txt_types:
            self._write_text_file()
            if not meta_comment == "":
                prepend_text_to_file(self.file_path, meta_comment)
        elif self.extension == '.yml':
            self._write_yaml_file()
            if not meta_comment == "":
                prepend_text_to_file(self.file_path, meta_comment,
                                     extra_line=True)
        else:
            raise ValueError("extension" +
                             "{} not supported".format(self.extension) +
                             ", supported extensions are (.yml|.csv|.txt)")


    def _write_text_file(self):
        """
        writer for .txt, .csv and .nk files
        """
        header = ""
        for key in Writer.DATA_META_DATA_KEYS:
            if key in self.file_dict['MetaData']:
                alias = Writer.KEY_ALIASES[key]
                header += "{}: {}\n".format(alias,
                                            self.file_dict['MetaData'][key])


        # for item in Writer.KEY_ALIASES.items():
        #     if item[0] in self.file_dict:
        #         header += "{}: {}\n".format(item[1], self.file_dict[item[0]])
        if len(self.file_dict['Datasets']) < 1:
            raise ValueError("no datasets present in the file dict")
        data_dict = self.file_dict["Datasets"][0]

        spec_type = multi_key(data_dict, {'spectrum_type', 'SpectrumType'})
        header += "SPECTRUMTYPE: {}\n".format(spec_type)
        unit = multi_key(data_dict, {'unit', 'Unit'})
        header += "UNIT: {}\n".format(unit)
        dtype = multi_key(data_dict, {'type', 'DataType'})
        header += "DATATYPE: {}\n".format(dtype)
        # dtype = multi_key(data_dict,{'type','DataType'})
        # if dtype not in {'tabulated_nk', 'tabulated_n'}:
        #     raise NotImplementedError("writing data to text file only"+
        #                               "supported for tabulated_nk and"+
        #                               "tabulted_n data.")

        if self.extension == '.txt':
            delimiter = "\t"
        elif self.extension == '.csv':
            delimiter = ","
        data = multi_key(data_dict, {'data', 'Data'})
        data = _str_table_to_numeric(data)
        np.savetxt(self.file_path, data,
                   delimiter=delimiter,
                   header=header,
                   fmt="%.8f")


    def _write_yaml_file(self):
        """
        writer for .yml files
        """

        pop_keys = []
        for key in self.file_dict:
            if self.file_dict[key] == "":
                pop_keys.append(key)
        for key in pop_keys:
            self.file_dict.pop(key)
        pop_dataset = []

        for ids, dataset in enumerate(self.file_dict['Datasets']):
            if ("constant" in dataset['DataType'] and
                    self.ignore_constant):
                pop_dataset.append(ids)
        for pop_id in pop_dataset:
            self.file_dict['Datasets'].pop(pop_id)
            # if isinstance(dataset['data'], np.ndarray):
            #     string_array = np.array2string(dataset['data'])
            #     no_brackets = re.sub(r'[\[\]]', '', string_array)
            #     no_indent = re.sub(r'\n ', r"\n", no_brackets)
            #     dataset['data'] = no_indent
            #if isinstance(dataset['coefficients'], )

        yaml_dict = {}
        for key in self.file_dict['MetaData']:
            if key in Writer.KEY_ALIASES:
                md = self.file_dict['MetaData'][key]
                yaml_dict[Writer.KEY_ALIASES[key]] = md
            else:
                yaml_dict[key] = md

        yaml_dict['DATA'] = self.file_dict['Datasets']


        if self.use_rii_aliases:
            yaml_datasets = []
            for dataset in yaml_dict['DATA']:

                yaml_dataset = {}
                for key in dataset:
                    if key in {"SpectrumType", "Unit", "Yields"}:
                        continue
                    value = dataset[key]
                    if value in Writer.RII_ALIASES:
                        value = Writer.RII_ALIASES[value]
                    if key in Writer.RII_ALIASES:
                        yaml_dataset[Writer.RII_ALIASES[key]] = value
                    else:
                        yaml_dataset[key] = value
                yaml_datasets.append(yaml_dataset)
            yaml_dict['DATA'] = yaml_datasets
        specs = yaml_dict.pop('SPECS')

        if len(specs) > 0:
            spec_order = ['n_absolute',
                          'wavelength_vacuum',
                          'film_thickness',
                          'substrate',
                          'temperature',
                          'pressure',
                          'deposition_temperature',
                          'direction']
            if USE_RUAMEL:
                yaml_spec = CommentedMap()
            else:
                yaml_spec = {}
            for spec_key in spec_order:
                for key, val in specs.items():
                    if spec_key == key:
                        yaml_spec[spec_key] = val
            yaml_dict['SPECS'] = yaml_spec

        write_yaml_file(self.file_path, yaml_dict)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="dispersion", # Replace with your own username
    version="1.0.4",
    author="Phillip Manley",
    author_email="manley@zib.de",
    description="support for libraries of optical dispersion (refractive index) data files",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nano-sippe/dispersion",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=['numpy', 'matplotlib', 'pandas', 'scipy', 'PyYAML'],
    python_requires='>=3.6',
    include_package_data=True,
    entry_points={
        'console_scripts': ['dispersion_setup='+
                            'dispersion.scripts.'+
                            'setup_dispersion:main',
                            'dispersion_catalogue_rebuild='+
                            'dispersion.scripts.'+
                            'catalogue_rebuild:main',],
    }
)

#data_files=[('config',['cfg/config.yaml'])],



"""describes spectrally dependent data

spectral_data implements the abstract base class SpectralData which
defines a material parameter which has a spectral dependence (e.g.
refractive index, permittivity). Each of the subclasses must implement
the evaluate method which returns the material parameter for a given
Spectrum object.

Classes
-------
SpectralData
    abstract base class
Constant: SpectralData
    for values that are independent of the spectrum
Interpolation: SpectralData
    for tabulated data values
Model: SpectralData
    abstract base class for values generated from a particular model
Sellmeier: Model
    implements the Sellmeier model for refractive index
Sellmeier2: Model
    implements the modified Sellmeier model for refractive index
Polynomial: Model
    implements a polynomial model for refractive index
RefractiveIndexInfo: Model
    implements the RefractiveIndexInfo model for refractive index
Cauchy: Model
    implements the Cauchy model for refractive index
Gases: Model
    implements the Gas model for refractive index
Herzberger: Model
    implements the Herzberger model for refractive index
Retro: Model
    implements the Retro model for refractive index
Exotic: Model
    implements the Exotic model for refractive index
Drude: Model
    implements the Drude model for complex permittivity
DrudeLorentz: Model
    implements the Drude-Lorentz model for complex permittivity
TaucLorentz: Model
    implements the Tauc-Lorentz model for complex permittivity

Notes
-----
for more information on models see https://refractiveindex.info/about
"""
import re
import numpy as np
from scipy.interpolate import interp1d, splrep, splev
from dispersion.spectrum import Spectrum
from dispersion.io import _numeric_to_string_table


class SpectralData():
    '''
    Base class for defining a quantity (e.g. refactive index)
    which is defined over a given spectrum (see class Spectrum).
    '''

    def __init__(self, valid_range,
                 spectrum_type='wavelength',
                 unit='nm'):
        self.spectrum_type = spectrum_type
        self.unit = unit
        self.valid_range = Spectrum(valid_range,
                                    spectrum_type=spectrum_type,
                                    unit=unit)

    def suggest_spectrum(self):
        """for plotting the spectral data we take a geometrically
        spaced set of values"""
        lower_bound = np.min(self.valid_range.values)
        upper_bound = np.max(self.valid_range.values)
        suggest = np.geomspace(lower_bound, upper_bound, num=1000)
        return Spectrum(suggest,
                        spectrum_type=self.spectrum_type,
                        unit=self.unit)

    def evaluate(self, spectrum):
        """returns the value of the spectral data for the given spectrum"""
        raise NotImplementedError("This method should be overridden by"+
                                  " a subclass")

class Constant(SpectralData):
    """for spectral data values that are independent of the spectrum"""

    def __init__(self, constant, valid_range=(0, np.inf),
                 spectrum_type='wavelength',
                 unit='m'):
        super(Constant, self).__init__(valid_range,
                                       spectrum_type=spectrum_type,
                                       unit=unit)
        self.constant = constant

    def evaluate(self, spectrum):
        """returns the value of the spectral data for the given spectrum"""
        self.valid_range.contains(spectrum)
        if isinstance(spectrum.values, (list, tuple, np.ndarray)):
            return self.constant * np.ones(len(spectrum.values))
        return self.constant

    def dict_repr(self):
        """
        return a dictionary representation of the object
        """
        data = {}
        data['DataType'] = "constant"
        data['ValidRange'] = _numeric_to_string_table(self.valid_range.values)
        data['SpectrumType'] = self.spectrum_type
        data['Unit'] = self.unit
        data['Value'] = self.constant
        return data

class Extrapolation(SpectralData):
    """
    for extending spectral data outside of the valid range.
    Use with caution
    """
    def __init__(self, spectral_data, extended_spectrum,
                 spline_order=2):
        self.base_spectral_data = spectral_data
        self.spline_order = spline_order
        extrap_spectrum = self.get_extrap_spectrum(extended_spectrum)
        min_range = np.min(extrap_spectrum.values)
        max_range = np.max(extrap_spectrum.values)
        spectrum_type = self.base_spectral_data.spectrum_type
        unit = self.base_spectral_data.unit
        super(Extrapolation, self).__init__((min_range, max_range),
                                            spectrum_type=spectrum_type,
                                            unit=unit)
        self.extrapolate_data()

    def get_extrap_spectrum(self,extended_spectrum):
        """
        takes a Spectrum object with one or two values possibly lying outside
        the base spectral range. Raises an error if the values do not lie
        outside the base spectral range. returns a new length two spectrum
        that gives the lower and upper bound for an extrapolation
        """
        base_spectrum = self.base_spectral_data.valid_range
        extended_spectrum.convert_to(spectrum_type= base_spectrum.spectrum_type,
                                     unit= base_spectrum.unit,
                                     in_place= True)
        new_range = np.array(base_spectrum.values)
        if isinstance(extended_spectrum.values, (list, tuple, np.ndarray)):
            # extrapolation both upper and lower
            extrap_values = extended_spectrum.values
            if extrap_values.size > 2:
                raise ValueError("extrapolation spectrum may contain at most" +
                                 "2 values not {}".format(extrap_values.size))
            for extrap_val in extrap_values:
                new_range = self.validate_extrap_val(extrap_val, new_range)
        else:
            # upper or lower
            extrap_val = extended_spectrum.values
            new_range = self.validate_extrap_val(extrap_val, new_range)

        return Spectrum(new_range,
                        spectrum_type= base_spectrum.spectrum_type,
                        unit= base_spectrum.unit)

    def validate_extrap_val(self,extrap_val,base_range):
        """
        checks if extrap_val lies outside base_range and replaces the relevant
        value in base_range with extrap_val.
        """
        if extrap_val < base_range[0]:
            base_range[0] = extrap_val
        elif extrap_val > base_range[1]:
            base_range[1] = extrap_val
        else:
            raise ValueError("extrapolation value of " +
                             "{} ".format(extrap_val) +
                             "lies inside the defined range " +
                             "{}".format(base_range) +
                             " therefore extrapolation is not necessary")
        return base_range


    def extrapolate_data(self):
        """makes a spline base on the base data for future lookup"""
        spectrum = self.base_spectral_data.suggest_spectrum()
        evaluation = self.base_spectral_data.evaluate(spectrum)
        self.extrapolation = splrep(spectrum.values, evaluation,
                                    k=self.spline_order)

    def evaluate(self, spectrum):
        """returns the value of the spectral data for the fiven spectrum"""
        try:
            self.base_spectral_data.valid_range.contains(spectrum)
            return self.base_spectral_data.evaluate(spectrum)
        except ValueError as e:
            spectrum.convert_to(self.spectrum_type, self.unit, in_place=True)
            self.valid_range.contains(spectrum)
            return splev(spectrum.values,self.extrapolation)


class Interpolation(SpectralData):
    """for spectral data values that are from tabulated data"""

    def __init__(self, data, spectrum_type='wavelength',
                 unit='m', interp_order=1):
        self.data = data
        self.interp_order = interp_order
        min_range = np.min(data[:, 0])
        max_range = np.max(data[:, 0])
        super(Interpolation, self).__init__((min_range, max_range),
                                            spectrum_type=spectrum_type,
                                            unit=unit)
        self.interpolate_data()

    def interpolate_data(self):
        """interpolates the data for future lookup"""
        self.interpolation = interp1d(self.data[:, 0], self.data[:, 1],
                                      kind=self.interp_order)

    def evaluate(self, spectrum):
        """returns the value of the spectral data for the fiven spectrum"""

        self.valid_range.contains(spectrum)
        values = spectrum.convert_to(self.spectrum_type, self.unit)
        return self.interpolation(values)

    def dict_repr(self):
        """
        return a dictionary representation of the object
        """
        data = {}
        data['DataType'] = "tabulated"
        data['ValidRange'] = _numeric_to_string_table(self.valid_range.values)
        data['SpectrumType'] = self.spectrum_type
        data['Unit'] = self.unit
        data['Data'] = _numeric_to_string_table(self.data)
        return data

class Model(SpectralData):
    """for spectral data values depending on model parameters"""

    def __init__(self, model_parameters, valid_range, spectrum_type='wavelength',
                 unit='m'):
        self.model_parameters = model_parameters
        super(Model, self).__init__(valid_range,
                                    spectrum_type=spectrum_type,
                                    unit=unit)
        self.required_spectrum_type = None # set in subclass
        self.required_unit = None # set in subclass
        self.output = None # set in subclass

    def validate_spectrum_type(self):
        tmp_spectrum = Spectrum(1.0,
                                spectrum_type=self.spectrum_type,
                                unit=self.unit)

        if not tmp_spectrum.spectrum_type == self.required_spectrum_type:
            raise ValueError("spectrum_type for model " +
                             "<{}> must".format(type(self).__name__) +
                             " be {}".format(self.required_spectrum_type))

        if not tmp_spectrum.unit == self.required_unit:
            raise ValueError("unit for model " +
                             "<{}> must".format(type(self).__name__) +
                             "be {}".format(self.required_unit))

    def dict_repr(self):
        """
        return a dictionary representation of the object
        """
        data = {}
        data['DataType'] = "model " + type(self).__name__
        data['ValidRange'] = _numeric_to_string_table(self.valid_range.values)
        data['SpectrumType'] = self.spectrum_type
        data['Unit'] = self.unit
        data['Yields'] = self.output
        data['Parameters'] = _numeric_to_string_table(self.model_parameters)
        return data

    def input_output(self):
        """defines the required inputs and the output spectrum type"""
        raise NotImplementedError("this abstract method needs to be"+
                                  "overridden by a subclass")

    def evaluate(self, spectrum):
        """returns the value of the spectral data for the given spectrum"""
        raise NotImplementedError("this abstract method needs to be"+
                                  "overridden by a subclass")

    def preprocess(self, spectrum):
        """
        check range of spectrum, convert to correct sType and unit and return
        an object with the same tensor order (scalar|vector) with values set
        to 1.0
        """
        self.valid_range.contains(spectrum)
        new_spectrum = spectrum.convert_to(self.spectrum_type,
                                           self.unit)
        if isinstance(spectrum.values, (list, tuple, np.ndarray)):
            ones = np.ones(new_spectrum.shape)
        else:
            ones = 1.0

        return ones, new_spectrum

class Sellmeier(Model):

    '''
    requires wavelength input in micrometers
    returns real part of refractive index only
    '''
    def __init__(self, model_parameters, valid_range,
                 spectrum_type='wavelength', unit='m'):
        super(Sellmeier, self).__init__(model_parameters, valid_range,
                                        spectrum_type=spectrum_type,
                                        unit=unit)
        self.required_spectrum_type = 'wavelength'
        self.required_unit = 'um'
        self.output = 'n'
        self.validate_spectrum_type()

    def evaluate(self, spectrum):
        """returns the value of the spectral data for the given spectrum"""
        [ones, wavelengths] = self.preprocess(spectrum)
        rhs = self.model_parameters[0]*ones
        wvlsq = np.power(wavelengths, 2)
        for iterc in range(len(self.model_parameters[1::2])):
            cupper = self.model_parameters[iterc*2+1]
            clower = self.model_parameters[iterc*2+2]
            rhs += cupper*wvlsq/(wvlsq-clower**2)
        ref_index = np.sqrt(rhs+1.0)
        return ref_index

class Sellmeier2(Model):

    '''
    requires wavelength input in micrometers
    returns real part of refractive index only
    '''
    def __init__(self, model_parameters, valid_range,
                 spectrum_type='wavelength', unit='m'):
        super(Sellmeier2, self).__init__(model_parameters, valid_range,
                                         spectrum_type=spectrum_type,
                                         unit=unit)
        self.required_spectrum_type = 'wavelength'
        self.required_unit = 'um'
        self.output = 'n'
        self.validate_spectrum_type()

    def evaluate(self, spectrum):
        """returns the value of the spectral data for the given spectrum"""
        [ones, wavelengths] = self.preprocess(spectrum)
        rhs = self.model_parameters[0]*ones
        wvlsq = np.power(wavelengths, 2)
        for iterc in range(len(self.model_parameters[1::2])):
            cupper = self.model_parameters[iterc*2+1]
            clower = self.model_parameters[iterc*2+2]
            rhs += cupper*wvlsq/(wvlsq-clower)
        ref_index = np.sqrt(rhs+1.0)
        return ref_index

class Polynomial(Model):

    '''
    requires wavelength input in micrometers
    returns real part of refractive index only
    '''
    def __init__(self, model_parameters, valid_range,
                 spectrum_type='wavelength', unit='m'):
        super(Polynomial, self).__init__(model_parameters, valid_range,
                                         spectrum_type=spectrum_type,
                                         unit=unit)
        self.required_spectrum_type = 'wavelength'
        self.required_unit = 'um'
        self.output = 'n'
        self.validate_spectrum_type()

    def evaluate(self, spectrum):
        """returns the value of the spectral data for the given spectrum"""
        [ones, wavelengths] = self.preprocess(spectrum)
        rhs = self.model_parameters[0]*ones

        for iterc in range(len(self.model_parameters[1::2])):
            c_multi = self.model_parameters[iterc*2+1]
            c_power = self.model_parameters[iterc*2+2]
            rhs += c_multi*np.power(wavelengths, c_power)
        ref_index = np.sqrt(rhs)
        return ref_index

class RefractiveIndexInfo(Model):

    '''
    requires wavelength input in micrometers
    returns real part of refractive index only
    '''
    def __init__(self, model_parameters, valid_range,
                 spectrum_type='wavelength', unit='m'):
        super(RefractiveIndexInfo, self).__init__(model_parameters, valid_range,
                                                  spectrum_type=spectrum_type,
                                                  unit=unit)
        self.required_spectrum_type = 'wavelength'
        self.required_unit = 'um'
        self.output = 'n'
        self.validate_spectrum_type()

    def evaluate(self, spectrum):
        """returns the value of the spectral data for the given spectrum"""
        [ones, wavelengths] = self.preprocess(spectrum)
        rhs = self.model_parameters[0]*ones
        wvlsq = np.power(wavelengths, 2)
        for iterc in range(len(self.model_parameters[1:8:4])):
            c_multi_upper = self.model_parameters[iterc*4+1]
            c_power_upper = self.model_parameters[iterc*4+2]
            c_multi_lower = self.model_parameters[iterc*4+3]
            c_power_lower = self.model_parameters[iterc*4+4]
            rhs += (c_multi_upper*np.power(wavelengths, c_power_upper)/
                    (wvlsq-np.power(c_multi_lower, c_power_lower)))
        for iterc in range(len(self.model_parameters[9::2])):
            c_multi = self.model_parameters[iterc*2+9]
            c_power = self.model_parameters[iterc*2+10]
            rhs += c_multi*np.power(wavelengths, c_power)
        ref_index = np.sqrt(rhs)
        return ref_index



class Cauchy(Model):

    '''
    requires wavelength input in micrometers
    returns real part of refractive index only
    '''
    def __init__(self, model_parameters, valid_range,
                 spectrum_type='wavelength', unit='m'):
        super(Cauchy, self).__init__(model_parameters, valid_range,
                                     spectrum_type=spectrum_type,
                                     unit=unit)
        self.required_spectrum_type = 'wavelength'
        self.required_unit = 'um'
        self.output = 'n'
        self.validate_spectrum_type()

    def evaluate(self, spectrum):
        """returns the value of the spectral data for the given spectrum"""
        [ones, wavelengths] = self.preprocess(spectrum)
        rhs = self.model_parameters[0]*ones
        for iterc in range(len(self.model_parameters[1::2])):
            c_multi = self.model_parameters[iterc*2+1]
            c_power = self.model_parameters[iterc*2+2]
            rhs += c_multi*np.power(wavelengths, c_power)
        ref_index = rhs
        return ref_index

class Gases(Model):

    '''
    requires wavelength input in micrometers
    returns real part of refractive index only
    '''
    def __init__(self, model_parameters, valid_range,
                 spectrum_type='wavelength', unit='m'):
        super(Gases, self).__init__(model_parameters, valid_range,
                                    spectrum_type=spectrum_type,
                                    unit=unit)
        self.required_spectrum_type = 'wavelength'
        self.required_unit = 'um'
        self.output = 'n'
        self.validate_spectrum_type()

    def evaluate(self, spectrum):
        """returns the value of the spectral data for the given spectrum"""
        [ones, wavelengths] = self.preprocess(spectrum)
        rhs = self.model_parameters[0]*ones
        wvlinvsq = np.power(wavelengths, -2)
        for iterc in range(len(self.model_parameters[1::2])):
            cupper = self.model_parameters[iterc*2+1]
            clower = self.model_parameters[iterc*2+2]
            rhs += cupper/(clower-wvlinvsq)
        ref_index = rhs+1.0
        return ref_index

class Herzberger(Model):

    '''
    requires wavelength input in micrometers
    returns real part of refractive index only
    '''
    def __init__(self, model_parameters, valid_range,
                 spectrum_type='wavelength', unit='m'):
        super(Herzberger, self).__init__(model_parameters, valid_range,
                                         spectrum_type=spectrum_type,
                                         unit=unit)
        self.required_spectrum_type = 'wavelength'
        self.required_unit = 'um'
        self.output = 'n'
        self.validate_spectrum_type()

    def evaluate(self, spectrum):
        """returns the value of the spectral data for the given spectrum"""
        [ones, wavelengths] = self.preprocess(spectrum)
        rhs = self.model_parameters[0]*ones
        wvlsq = np.power(wavelengths, 2)
        rhs += self.model_parameters[1]*np.power(wvlsq-0.028, -1)
        rhs += self.model_parameters[2]*np.power(wvlsq-0.028, -2)
        rhs += self.model_parameters[3]*wvlsq
        rhs += self.model_parameters[4]*np.power(wavelengths, 4)
        rhs += self.model_parameters[5]*np.power(wavelengths, 6)

        ref_index = rhs
        return ref_index

class Retro(Model):

    '''
    requires wavelength input in micrometers
    returns real part of refractive index only
    '''
    def __init__(self, model_parameters, valid_range,
                 spectrum_type='wavelength', unit='m'):
        super(Retro, self).__init__(model_parameters, valid_range,
                                    spectrum_type=spectrum_type,
                                    unit=unit)
        self.required_spectrum_type = 'wavelength'
        self.required_unit = 'um'
        self.output = 'n'
        self.validate_spectrum_type()

    def evaluate(self, spectrum):
        """returns the value of the spectral data for the given spectrum"""
        [ones, wavelengths] = self.preprocess(spectrum)
        rhs = self.model_parameters[0]*ones
        wvlsq = np.power(wavelengths, 2)
        rhs += self.model_parameters[1]*wvlsq/(wvlsq-self.model_parameters[2])
        rhs += self.model_parameters[3]*wvlsq

        tmp_p = -2*rhs/(1-rhs)
        tmp_q = -1/(1-rhs)

        ref_index = -0.5*tmp_p + np.sqrt(np.power(0.5*tmp_p, 2) - tmp_q)
        return ref_index

class Exotic(Model):

    '''
    requires wavelength input in micrometers
    returns real part of refractive index only
    '''
    def __init__(self, model_parameters, valid_range,
                 spectrum_type='wavelength', unit='m'):
        super(Exotic, self).__init__(model_parameters, valid_range,
                                     spectrum_type=spectrum_type,
                                     unit=unit)
        self.required_spectrum_type = 'wavelength'
        self.required_unit = 'um'
        self.output = 'n'
        self.validate_spectrum_type()

    def evaluate(self, spectrum):
        """returns the value of the spectral data for the given spectrum"""
        [ones, wavelengths] = self.preprocess(spectrum)
        rhs = self.model_parameters[0]*ones
        wvlsq = np.power(wavelengths, 2)

        rhs += self.model_parameters[1]*wvlsq/(wvlsq - self.model_parameters[2])
        rhs += (self.model_parameters[3]*(wavelengths -self.model_parameters[4])/
                (np.power(wavelengths - self.model_parameters[4], 2) +
                 self.model_parameters[5]))
        ref_index = np.sqrt(rhs)
        return ref_index

class Drude(Model):

    '''
    requires energy input in eV
    returns real and imaginary parts of permittivity
    '''
    def __init__(self, model_parameters, valid_range,
                 spectrum_type='wavelength', unit='m'):
        super(Drude, self).__init__(model_parameters, valid_range,
                                    spectrum_type=spectrum_type,
                                    unit=unit)
        self.required_spectrum_type = 'energy'
        self.required_unit = 'ev'
        self.output = 'eps'
        self.validate_spectrum_type()

    def evaluate(self, spectrum):
        """returns the value of the spectral data for the given spectrum"""
        [ones, energies] = self.preprocess(spectrum)
        omega_p = self.model_parameters[0] # plasma frequency in eV
        loss = self.model_parameters[1] # loss in eV
        return ones - omega_p**2/(np.power(energies, 2)+1j*loss*energies)

class DrudeLorentz(Model):

    '''
    requires energy input in eV
    returns real and imaginary parts of permittivity
    '''
    def __init__(self, model_parameters, valid_range,
                 spectrum_type='wavelength', unit='m'):
        super(DrudeLorentz, self).__init__(model_parameters, valid_range,
                                           spectrum_type=spectrum_type,
                                           unit=unit)
        self.required_spectrum_type = 'energy'
        self.required_unit = 'ev'
        self.output = 'eps'
        self.validate_spectrum_type()

    def evaluate(self, spectrum):
        """returns the value of the spectral data for the given spectrum"""
        [ones, energies] = self.preprocess(spectrum)
        omega_p = self.model_parameters[0] # plasma frequency in eV
        pol_str = self.model_parameters[1] # pole strength (0.<#<1.)
        w_res = self.model_parameters[2] # frequency of Lorentz pole in eV
        loss = self.model_parameters[3] # loss in eV
        return ones +  np.conj(pol_str*omega_p**2/((w_res**2-np.power(energies, 2))+1j*loss*energies))

class TaucLorentz(Model):
    '''
    requires energy input in eV
    returns real and imaginary parts of permittivity
    '''
    def __init__(self, model_parameters, valid_range,
                 spectrum_type='wavelength', unit='m'):
        super(TaucLorentz, self).__init__(model_parameters, valid_range,
                                           spectrum_type=spectrum_type,
                                           unit=unit)
        self.required_spectrum_type = 'energy'
        self.required_unit = 'ev'
        self.output = 'eps'
        self.validate_spectrum_type()

    def evaluate(self, spectrum):
        """returns the value of the spectral data for the given spectrum"""
        [ones, energies] = self.preprocess(spectrum)
        A = self.model_parameters[0] # oscillator strength
        E0 = self.model_parameters[1] # pole energy
        C = self.model_parameters[2] # pole broadening
        Eg = self.model_parameters[3] # optical bandgap energy
        eps_inf = self.model_parameters[4] # high frequency limit of the real part of permittivity
        eps_imag = self._calc_eps_imag(ones, energies, A, E0, C, Eg)
        eps_real = self._calc_eps_real(energies, A, E0, C, Eg, eps_inf)
        eps = eps_real + 1j*eps_imag
        return eps

    def _calc_eps_imag(self, ones, energies, A, E0, C, Eg):
        eps_imag = ones
        E = energies[energies>=Eg]
        eps_imag[energies>=Eg] = ( (1./E) *A*E0*C*(E-Eg)**2 /
                                   ((E**2-E0**2)**2+C**2*E**2))
        eps_imag[energies<Eg] = 0.0
        return eps_imag

    def _calc_eps_real(self, energies, A, E0, C, Eg, eps_inf):
        E = energies
        alpha_ln = ((Eg**2-E0**2)*E**2 +
                    Eg**2*C**2 -
                    E0**2*(E0**2+3*Eg**2))
        alpha_atan = (E**2-E0**2)*(E0**2+Eg**2) + Eg**2*C**2
        alpha = np.sqrt(4*E0**2-C**2)
        gamma = np.sqrt(E0**2 -0.5*C**2)
        zeta4 = (E**2-gamma**2)**2 + 0.25*alpha**2*C**2

        part1 = (0.5*(A*C*alpha_ln/(np.pi*zeta4*alpha*E0)) *
                 np.log( (E0**2+Eg**2+alpha*Eg)/(E0**2+Eg**2-alpha*Eg)))

        part2 = ((-1*A*alpha_atan/(np.pi*zeta4*E0)) *
                 (np.pi - np.arctan( (2*Eg+alpha)/C) +
                  np.arctan( (-2*Eg+alpha)/C)))

        # From original paper, seems to be wrong
        # part3 = ((2.*A*E0*C/(np.pi*zeta4)) *
        #          (Eg*(E**2-gamma**2)*
        #           (np.pi+2*np.arctan2((gamma**2-Eg**2),(alpha*C)))))

        part3_1 = (4.*A*E0/(np.pi*zeta4*alpha))
        part3_2 = Eg*(E**2-gamma**2)
        part3_3 = ( np.arctan2(alpha+2*Eg,C) +
                    np.arctan2(alpha-2*Eg,C))
        part3 = part3_1 * part3_2 * part3_3

        part4 = ((-A*E0*C/(np.pi*zeta4))*
                 ((E**2+Eg**2)/E) *
                 np.log( np.abs(E-Eg)/(E+Eg)))

        part5 = ((2.*A*E0*C*Eg/(np.pi*zeta4)) *
                 np.log( (np.abs(E-Eg)*(E+Eg))/
                         np.sqrt((E0**2-Eg**2)**2+Eg**2*C**2)))

        eps_real = eps_inf + part1 + part2 + part3 + part4 +part5
        return eps_real



class Fano(Model):

    '''
    this model can be applied to scattering cross sections
    requires energy input in eV
    returns real and imaginary parts of scattering cross section
    '''
    def input_output(self):
        """defines the required inputs and the output spectrum type"""
        self.required_spectrum_type = 'energy'
        self.required_unit = 'ev'
        self.output = 'scattering_cross_setion'

    def evaluate(self, spectrum):
        """returns the value of the spectral data for the given spectrum"""
        [ones, energies] = self.preprocess(spectrum)
        q = self.model_parameters[0] # Fano parameter
        e_r = self.model_parameters[1] # resonant energy in eV
        gamma = self.model_parameters[2] # loss in eV
        epsilon =  2*(energies-e_r)/gamma
        #loss = self.model_parameters[1] # loss in eV
        norm = (1+q**2)
        sigma = (1/norm) *(epsilon+q)**2 / (epsilon**2+1)
        return sigma
